'''
Hyrule Castle
'''

__all__ = 'LOCATIONS',


LOCATIONS = {
    'Hyrule Castle': {
        'type': 'area',
        'link': {
            'Pyramid': [('and', [
                ('settings', 'inverted'), ('item', 'mirror')])],
            'North Light World': [],
            'South Light World': [],
            'East Light World': [('item', 'powerglove')],
            'Secret Path Drop Entrance (E)': [('rabbitbarrier', None)],
            'Gate Portal': [('and', [
                ('access', 'Castle Tower Reward'),
                ('nosettings', 'inverted')])],
            'Secret Path Exit Entrance (E)': [('rabbitbarrier', None)],
            'Castle Gates Entrance (E)': []}
    },
    'Gate Portal': {
        'type': 'area',
        'link': {
            'Pyramid': [('state', 'rabbit')]}
    },
    'Secret Path Drop Entrance (E)': {
        'type': 'entrance_drop', 'map': 'light', 'coord': (395, 276),
        'link': {
            'Secret Path Drop Entrance (I)': []}
    },
    'Secret Path Drop Entrance (I)': {
        'type': 'interior',
        'link': {
            'Secret Path': []}
    },
    'Secret Path': {
        'type': 'drop', 'map': 'light', 'coord': (365, 284),
        'link': {
            'Secret Path Exit Entrance (I)': []}
    },
    'Secret Path Exit Entrance (I)': {
        'type': 'interior',
        'link': {
            'Secret Path': [],
            'Secret Path Exit Entrance (E)': []}
    },
    'Secret Path Exit Entrance (E)': {
        'type': 'entrance_unique', 'map': 'light', 'coord': (365, 284),
        'link': {
            'Secret Path Exit Entrance (I)': [],
            'Hyrule Castle': []}
    },

    'Castle Gates Entrance (E)': {
        'type': 'entrance_unique', 'map': 'light', 'coord': (332, 291),
        'link': {
            'Hyrule Castle': [],
            'Castle Gates Entrance (I)': []}
    },
    'Castle Gates Entrance (I)': {
        'type': 'interior',
        'link': {
            'Castle Gates Entrance (E)': [],
            'Hyrule Castle Interior': []}
    },
    'Hyrule Castle Interior': {
        'type': 'area', 'dungeon': 'Castle Dungeons',
        'link': {
            'Castle Gates Entrance (I)': [],
            'Western Wall Access Entrance (I)': [],
            'Eastern Wall Access Entrance (I)': [],
            'Secret Escape': [
                ('and', [('access', 'Zelda'), ('rabbitbarrier', None)]),
                ('nosettings', 'standard')],
            'Dungeons Descent 1': []}
    },
    'Dungeons Descent 1': {
        'type': 'area', 'dungeon': 'Castle Dungeons',
        'link': {
            'Hyrule Castle Interior': [],
            'Dungeons Map': [],
            'Dungeons First Soldier Key': [('rabbitbarrier', None)],
            'Dungeons Descent 2': [('smallkey', 'Castle Dungeons')]}
    },
    'Dungeons Map': {
        'type': 'dungeonchest', 'dungeon': 'Castle Dungeons',
        'link': {
            'Dungeons Descent 1': []}
    },
    'Dungeons First Soldier Key': {
        'type': 'dungeonkey', 'dungeon': 'Castle Dungeons',
        'link': {
            'Dungeons Descent 1': []}
    },
    'Dungeons Descent 2': {
        'type': 'area', 'dungeon': 'Castle Dungeons',
        'link': {
            'Dungeons Descent 1': [],
            'Dungeons Second Soldier Key': [],
            'Dungeons Treasure': [],
            'Dungeons Cells': [('smallkey', 'Castle Dungeons')]}
    },
    'Dungeons Second Soldier Key': {
        'type': 'dungeonkey', 'dungeon': 'Castle Dungeons',
        'link': {
            'Dungeons Descent 2': []}
    },
    'Dungeons Treasure': {
        'type': 'dungeonchest', 'dungeon': 'Castle Dungeons',
        'link': {
            'Dungeons Descent 2': []}
    },
    'Dungeons Cells': {
        'type': 'area', 'dungeon': 'Castle Dungeons',
        'link': {
            'Dungeons Descent 2': [],
            'Dungeons Big Key': []}
    },
    'Dungeons Big Key': {
        'type': 'area', 'dungeon': 'Castle Dungeons',
        'link': {
            'Dungeons Cells': [],
            'Zelda': []}
    },
    'Zelda': {
        'type': 'area', 'dungeon': 'Castle Dungeons',
        'link': {
            'Dungeons Cells': [],
            "Zelda's Treasure": []}
    },
    "Zelda's Treasure": {
        'type': 'dungeonchest', 'dungeon': 'Castle Dungeons',
        'link': {
            'Zelda': [],
            'Castle Dungeons': [('rabbitbarrier', None)]}
    },

    'Secret Escape': {
        'type': 'area', 'dungeon': 'Castle Dungeons',
        'link': {
            'Hyrule Castle Interior': [],
            'Sewers': [
                ('item', 'lantern'), ('settings', 'standard'),
                ('and', [
                    ('settings', 'placement_advanced'), ('item', 'firerod')])]}
    },
    'Sewers': {
        'type': 'area', 'dungeon': 'Castle Dungeons',
        'link': {
            'Sewers Chest': [('rabbitbarrier', None)],
            'Sewers Rat Key': [],
            'Sewers Locked Door': [('smallkey', 'Castle Dungeons')]}
    },
    'Sewers Chest': {
        'type': 'drop', 'map': 'light', 'coord': (344, 223),
        'link': {
            'Sewers': [],
            'Sewers Dungeon Chest': []}
    },
    'Sewers Dungeon Chest': {
        'type': 'dungeonchest', 'dungeon': 'Castle Dungeons',
        'link': {
            'Sewers Chest': []}
    },
    'Sewers Rat Key': {
        'type': 'dungeonkey', 'dungeon': 'Castle Dungeons',
        'link': {
            'Sewers': []}
    },
    'Sewers Locked Door': {
        'type': 'area', 'dungeon': 'Castle Dungeons',
        'link': {
            'Sewers': [
                ('item', 'lantern'), ('settings', 'standard'),
                ('and', [
                    ('settings', 'placement_advanced'), ('item', 'firerod')])],
            'Sewers Last Hall': [('state', 'maybe;add')]}
    },
    'Sewers Last Hall': {
        'type': 'area', 'dungeon': 'Castle Dungeons',
        'link': {
            'Sewers Locked Door': [('state', 'maybe;add')],
            'Sewers Treasure': [('item', 'bombs'), ('item', 'pegasus')],
            'Sanctuary': [('rabbitbarrier', None)]}
    },
    'Sewers Treasure': {
        'type': 'drop', 'map': 'light', 'coord': (344, 193),
        'link': {
            'Sewers Last Hall': [],
            'Sewers Treasure Chest 1': [],
            'Sewers Treasure Chest 2': [],
            'Sewers Treasure Chest 3': []}
    },
    'Sewers Treasure Chest 1': {
        'type': 'dungeonchest', 'dungeon': 'Castle Dungeons',
        'link': {
            'Sewers Treasure': []}
    },
    'Sewers Treasure Chest 2': {
        'type': 'dungeonchest', 'dungeon': 'Castle Dungeons',
        'link': {
            'Sewers Treasure': []}
    },
    'Sewers Treasure Chest 3': {
        'type': 'dungeonchest', 'dungeon': 'Castle Dungeons',
        'link': {
            'Sewers Treasure': []}
    },
    'Sanctuary': {
        'type': 'chest', 'map': 'light', 'coord': (306, 172),
        'link': {
            'Sanctuary Entrance Entrance (I)': []}
    },

    'Western Wall Access Entrance (I)': {
        'type': 'interior',
        'link': {
            'Hyrule Castle Interior': [],
            'Western Wall Access Entrance (E)': []}
    },
    'Eastern Wall Access Entrance (I)': {
        'type': 'interior',
        'link': {
            'Hyrule Castle Interior': [],
            'Eastern Wall Access Entrance (E)': []}
    },
    'Western Wall Access Entrance (E)': {
        'type': 'entrance_unique', 'map': 'light', 'coord': (298, 257),
        'link': {
            'Western Wall Access Entrance (I)': [],
            'Castle Walls': []}
    },
    'Eastern Wall Access Entrance (E)': {
        'type': 'entrance_unique', 'map': 'light', 'coord': (365, 257),
        'link': {
            'Eastern Wall Access Entrance (I)': [],
            'Castle Walls': []}
    },
    'Castle Walls': {
        'type': 'area',
        'link': {
            'Pyramid': [('and', [
                ('settings', 'inverted'), ('item', 'mirror')])],
            'Western Wall Access Entrance (E)': [],
            'Eastern Wall Access Entrance (E)': [],
            'Hyrule Castle': [],
            'Castle Tower Entrance (E)': [
                ('and', [
                    ('nosettings', 'inverted'),
                    ('or', [
                        ('item', 'mastersword'), ('item', 'cape'),
                        ('and', [
                            ('settings', 'swordless'),
                            ('item', 'hammer')])])]),
                ('and', [
                    ('settings', 'inverted'), ('macro', 'ganonstower')])],
            'Ganon Drop Entrance (E)': [('and', [
                ('settings', 'inverted'),
                ('or', [
                    ('and', [
                        ('nosettings', 'entrance'), ('settings', 'fastganon')]),
                    ('macro', 'ganondrop')])])]}
    },
    'Castle Tower Entrance (E)': {
        'type': 'entrance_dungeon', 'map': 'light', 'coord': (332, 265),
        'link': {
            'Castle Walls': [],
            'Castle Tower Entrance (I)': [('nosettings', 'inverted')],
            "Ganon's Tower Entrance (I)": [('settings', 'inverted')]}
    },
    'Castle Tower Entrance (I)': {
        'type': 'interior',
        'link': {
            'Castle Tower Entrance (E)': [('nosettings', 'inverted')],
            "Ganon's Tower Entrance (E)": [('settings', 'inverted')],
            'Castle Tower Ascent 1': []}
    },
    'Castle Tower Ascent 1': {
        'type': 'area', 'dungeon': 'Castle Tower',
        'link': {
            'Castle Tower Entrance (I)': [],
            'Castle Tower First Chest': [],
            'Castle Tower Ascent Stairs': [('smallkey', 'Castle Tower')]}
    },
    'Castle Tower First Chest': {
        'type': 'dungeonchest', 'dungeon': 'Castle Tower',
        'link': {
            'Castle Tower Ascent 1': []}
    },
    'Castle Tower Ascent Stairs': {
        'type': 'area', 'dungeon': 'Castle Tower',
        'link': {
            'Castle Tower First Chest': [],
            'Castle Tower Ascent 2': [
                ('item', 'lantern'),
                ('and', [
                    ('settings', 'placement_advanced'), ('item', 'firerod')])]}
    },
    'Castle Tower Ascent 2': {
        'type': 'area', 'dungeon': 'Castle Tower',
        'link': {
            'Castle Tower Ascent Stairs': [],
            'Castle Tower Second Chest': [],
            'Castle Tower Ascent 3': [('smallkey', 'Castle Tower')]}
    },
    'Castle Tower Second Chest': {
        'type': 'dungeonchest', 'dungeon': 'Castle Tower',
        'link': {
            'Castle Tower Ascent 2': []}
    },
    'Castle Tower Ascent 3': {
        'type': 'area', 'dungeon': 'Castle Tower',
        'link': {
            'Castle Tower Ascent 2': [
                ('item', 'lantern'),
                ('and', [
                    ('settings', 'placement_advanced'), ('item', 'firerod')])],
            'Castle Tower Boss': [('item', 'sword'), ('settings', 'swordless')]}
    },
    'Castle Tower Boss': {
        'type': 'dungeonboss', 'dungeon': 'Castle Tower',
        'link': {
            'Castle Tower Boss Item': [
                ('item', 'sword'),
                ('and', [
                    ('settings', 'swordless'), ('item', 'hammer')]),
                ('item', 'bugnet')]}
    },
    'Castle Tower Boss Item': {
        'type': 'area', 'dungeon': 'Castle Tower',
        'link': {
            'Castle Tower Reward': [('boss', 'Castle Tower')]}
    },
    'Castle Tower Reward': {
        'type': 'area', 'dungeon': 'Castle Tower',
        'link': {
            'Pyramid': [('and', [
                ('nosettings', 'inverted'), ('state', 'rabbit')])],
            'Castle Walls': [('and', [
                ('settings', 'inverted'), ('state', 'rabbit')])]}
    },
}



'''
Access point to program GUI.
'''

import importlib

from .config import CONFIG

guilib = importlib.import_module(
    '.{0:s}'.format(CONFIG['gui']), package=__package__)

__all__ = 'guilib', 'GUI',

GUI = None



'''
Turtle Rock
'''

__all__ = 'LOCATIONS',


LOCATIONS = {
    'Turtle Rock Entrance Entrance (E)': {
        'type': 'entrance_dungeon', 'map': 'dark', 'coord': (618, 54),
        'link': {
            'Turtle Rock Entrance Entrance (I)': [
                ('settings', 'placement_advanced'),
                ('and', [
                    ('or', [
                        ('settings', 'swordless'), ('item', 'mastersword')]),
                    ('or', [
                        ('settings', 'bottle'), ('item', 'bluemail')])])]}
    },
    'Turtle Rock Entrance Entrance (I)': {
        'type': 'interior',
        'link': {
            'Turtle Rock Entrance Entrance (E)': [],
            'Turtle Rock Big Hole': [('item', 'somaria')]}
    },
    'Turtle Rock Big Hole': {
        'type': 'area', 'dungeon': 'Turtle Rock',
        'link': {
            'Turtle Rock Entrance Entrance (I)': [],
            'Turtle Rock Compass': [],
            'Turtle Rock Rolling Spikes': [('and', [
                ('item', 'somaria'), ('item', 'firerod')])],
            'Turtle Rock Trapped Stalfos': []}
    },
    'Turtle Rock Compass': {
        'type': 'dungeonchest', 'dungeon': 'Turtle Rock',
        'link': {
            'Turtle Rock Big Hole': [('item', 'somaria')]}
    },
    'Turtle Rock Rolling Spikes': {
        'type': 'area', 'dungeon': 'Turtle Rock',
        'link': {
            'Turtle Rock Big Hole': [('item', 'somaria')],
            'Turtle Rock Map': [],
            'Turtle Rock Map Key': []}
    },
    'Turtle Rock Map': {
        'type': 'dungeonchest', 'dungeon': 'Turtle Rock',
        'link': {
            'Turtle Rock Rolling Spikes': []}
    },
    'Turtle Rock Map Key': {
        'type': 'dungeonchest', 'dungeon': 'Turtle Rock',
        'link': {
            'Turtle Rock Rolling Spikes': []}
    },
    'Turtle Rock Trapped Stalfos': {
        'type': 'area', 'dungeon': 'Turtle Rock',
        'link': {
            'Turtle Rock Big Hole': [('item', 'somaria')],
            'Turtle Rock Hokkubokku Key Room': [('smallkey', 'Turtle Rock')]},
    },
    'Turtle Rock Hokkubokku Key Room': {
        'type': 'area', 'dungeon': 'Turtle Rock',
        'link': {
            'Turtle Rock Trapped Stalfos': [('smallkey', 'Turtle Rock')],
            'Turtle Rock Chain Chomp Room': [('smallkey', 'Turtle Rock')],
            'Turtle Rock First Hokkubokku Key': []}
    },
    'Turtle Rock First Hokkubokku Key': {
        'type': 'dungeonkey', 'dungeon': 'Turtle Rock',
        'link': {
            'Turtle Rock Hokkubokku Key Room': []}
    },
    'Turtle Rock Chain Chomp Room': {
        'type': 'area', 'dungeon': 'Turtle Rock',
        'link': {
            'Turtle Rock Hokkubokku Key Room': [('smallkey', 'Turtle Rock')],
            'Turtle Rock Chain Chomp Key': [],
            'Turtle Rock Pipe Rooms': [('smallkey', 'Turtle Rock')]}
    },
    'Turtle Rock Chain Chomp Key': {
        'type': 'dungeonchest', 'dungeon': 'Turtle Rock',
        'link': {
            'Turtle Rock Chain Chomp Room': []}
    },
    'Turtle Rock Pipe Rooms': {
        'type': 'area', 'dungeon': 'Turtle Rock',
        'link': {
            'Turtle Rock Chain Chomp Room': [('smallkey', 'Turtle Rock')],
            'Turtle Rock Second Hokkubokku Key': [],
            'Turtle Rock Big Key': [('smallkey', 'Turtle Rock')],
            'Turtle Rock Hokkubokku Room': []}
    },
    'Turtle Rock Second Hokkubokku Key': {
        'type': 'dungeonkey', 'dungeon': 'Turtle Rock',
        'link': {
            'Turtle Rock Pipe Rooms': []}
    },
    'Turtle Rock Big Key': {
        'type': 'dungeonchest', 'dungeon': 'Turtle Rock',
        'link': {
            'Turtle Rock Pipe Rooms': []}
    },
    'Turtle Rock Hokkubokku Room': {
        'type': 'area', 'dungeon': 'Turtle Rock',
        'link': {
            'Turtle Rock Pipe Rooms': [],
            'Turtle Rock Laser Eye Exit Entrance (I)': [],
            'Turtle Rock Zoro Room': [('bigkey', 'Turtle Rock')]}
    },
    'Turtle Rock Laser Eye Exit Entrance (I)': {
        'type': 'interior',
        'link': {
            'Turtle Rock Hokkubokku Room': [('and', [
                ('rabbitbarrier', None),
                ('or', [
                    ('settings', 'placement_advanced'),
                    ('and', [
                        ('or', [
                            ('settings', 'swordless'),
                            ('item', 'mastersword')]),
                        ('or', [
                            ('settings', 'bottle'),
                            ('item', 'bluemail')])])])])],
            'Turtle Rock Laser Eye Exit Entrance (E)': []}
    },
    'Turtle Rock Laser Eye Exit Entrance (E)': {
        'type': 'entrance_dungeon', 'map': 'dark', 'coord': (522, 60),
        'link': {
            'Turtle Rock Laser Eye Exit Entrance (I)': [],
            'Turtle Rock Balcony': []}
    },
    'Turtle Rock Balcony': {
        'type': 'area',
        'link': {
            'Spiral Cave Top Entrance Entrance (E)': [('and', [
                ('nosettings', 'inverted'), ('item', 'mirror')])],
            'Goriya Cave Entrance (E)': [('and', [
                ('nosettings', 'inverted'), ('item', 'mirror')])],
            'Turtle Rock Laser Eye Exit Entrance (E)': [],
            'Turtle Rock Big Chest Exit Entrance (E)': []}
    },
    'Turtle Rock Big Chest Exit Entrance (E)': {
        'type': 'entrance_dungeon', 'map': 'dark', 'coord': (553, 60),
        'link': {
            'Turtle Rock Balcony': [],
            'Turtle Rock Big Chest Exit Entrance (I)': []}
    },
    'Turtle Rock Big Chest Exit Entrance (I)': {
        'type': 'interior',
        'link': {
            'Turtle Rock Big Chest Exit Entrance (E)': [],
            'Turtle Rock Treasure': [('and', [
                ('or', [
                    ('item', 'hookshot'), ('item', 'somaria')]),
                ('or', [
                    ('settings', 'placement_advanced'),
                    ('and', [
                        ('or', [
                            ('settings', 'swordless'),
                            ('item', 'mastersword')]),
                        ('or', [
                            ('settings', 'bottle'),
                            ('item', 'bluemail')])])])])],
            'Turtle Rock Hokkubokku Room': [('and', [
                ('item', 'pegasus'),
                ('or', [
                    ('settings', 'placement_advanced'),
                    ('and', [
                        ('or', [
                            ('settings', 'swordless'),
                            ('item', 'mastersword')]),
                        ('or', [
                            ('settings', 'bottle'),
                            ('item', 'bluemail')])])])])]}
    },
    'Turtle Rock Treasure': {
        'type': 'dungeonchest_nokey', 'dungeon': 'Turtle Rock',
        'link': {
            'Turtle Rock Big Chest Exit Entrance (I)': [('item', 'somaria')],
            'Turtle Rock Hokkubokku Room': []}
    },
    'Turtle Rock Zoro Room': {
        'type': 'area', 'dungeon': 'Turtle Rock',
        'link': {
            'Turtle Rock Hokkubokku Room': [],
            'Turtle Rock Switch Room': [('item', 'bombs'), ('item', 'pegasus')]}
    },
    'Turtle Rock Switch Room': {
        'type': 'area', 'dungeon': 'Turtle Rock',
        'link': {
            'Turtle Rock Zoro Room': [],
            'Turtle Rock Switch Key': [],
            'Turtle Rock Dark Room North': [('smallkey', 'Turtle Rock')]}
    },
    'Turtle Rock Switch Key': {
        'type': 'dungeonchest', 'dungeon': 'Turtle Rock',
        'link': {
            'Turtle Rock Switch Room': []}
    },
    'Turtle Rock Dark Room North': {
        'type': 'area', 'dungeon': 'Turtle Rock',
        'link': {
            'Turtle Rock Switch Room': [('smallkey', 'Turtle Rock')],
            'Turtle Rock Dark Room Maze': [('and', [
                ('item', 'lantern'), ('item', 'somaria')])]}
    },
    'Turtle Rock Dark Room Maze': {
        'type': 'area', 'dungeon': 'Turtle Rock',
        'link': {
            'Turtle Rock Dark Room North': [],
            'Turtle Rock Dark Room South': []}
    },
    'Turtle Rock Dark Room South': {
        'type': 'area', 'dungeon': 'Turtle Rock',
        'link': {
            'Turtle Rock Dark Room Maze': [('and', [
                ('item', 'lantern'), ('item', 'somaria')])],
            'Turtle Rock Treasury': []}
    },
    'Turtle Rock Treasury': {
        'type': 'area', 'dungeon': 'Turtle Rock',
        'link': {
            'Turtle Rock Dark Room South': [],
            'Turtle Rock Treasury Chest 1': [
                ('item', 'mirrorshield'), ('item', 'byrna'), ('item', 'cape'),
                ('settings', 'placement_advanced')],
            'Turtle Rock Treasury Chest 2': [
                ('item', 'mirrorshield'), ('item', 'byrna'), ('item', 'cape'),
                ('settings', 'placement_advanced')],
            'Turtle Rock Treasury Chest 3': [
                ('item', 'mirrorshield'), ('item', 'byrna'), ('item', 'cape'),
                ('settings', 'placement_advanced')],
            'Turtle Rock Treasury Chest 4': [
                ('item', 'mirrorshield'), ('item', 'byrna'), ('item', 'cape'),
                ('settings', 'placement_advanced')],
            'Turtle Rock Fairy Exit Entrance (I)': [('item', 'bombs')],
            'Turtle Rock Boss Maze': [('smallkey', 'Turtle Rock')]}
    },
    'Turtle Rock Treasury Chest 1': {
        'type': 'dungeonchest', 'dungeon': 'Turtle Rock',
        'link': {
            'Turtle Rock Treasury': []}
    },
    'Turtle Rock Treasury Chest 2': {
        'type': 'dungeonchest', 'dungeon': 'Turtle Rock',
        'link': {
            'Turtle Rock Treasury': []}
    },
    'Turtle Rock Treasury Chest 3': {
        'type': 'dungeonchest', 'dungeon': 'Turtle Rock',
        'link': {
            'Turtle Rock Treasury': []}
    },
    'Turtle Rock Treasury Chest 4': {
        'type': 'dungeonchest', 'dungeon': 'Turtle Rock',
        'link': {
            'Turtle Rock Treasury': []}
    },
    'Turtle Rock Fairy Exit Entrance (I)': {
        'type': 'interior',
        'link': {
            'Turtle Rock Treasury': [('and', [
                ('rabbitbarrier', None),
                ('or', [
                    ('settings', 'placement_advanced'),
                    ('and', [
                        ('or', [
                            ('settings', 'swordless'),
                            ('item', 'mastersword')]),
                        ('or', [
                            ('settings', 'bottle'),
                            ('item', 'bluemail')])])])])],
            'Turtle Rock Fairy Exit Entrance (E)': []}
    },
    'Turtle Rock Fairy Exit Entrance (E)': {
        'type': 'entrance_dungeon', 'map': 'dark', 'coord': (538, 76),
        'link': {
            'Turtle Rock Fairy Exit Entrance (I)': [],
            'Turtle Rock Fairy Exit': []}
    },
    'Turtle Rock Fairy Exit': {
        'type': 'area',
        'link': {
            'Fairy Maze Top Entrance Entrance (E)': [('and', [
                ('nosettings', 'inverted'), ('item', 'mirror')])],
            'Turtle Rock Fairy Exit Entrance (E)': []}
    },
    'Turtle Rock Boss Maze': {
        'type': 'area', 'dungeon': 'Turtle Rock',
        'link': {
            'Turtle Rock Treasury': [],
            'Turtle Rock Boss Door': [('item', 'somaria')]}
    },
    'Turtle Rock Boss Door': {
        'type': 'area', 'dungeon': 'Turtle Rock',
        'link': {
            'Turtle Rock Boss Maze': [('item', 'somaria')],
            'Turtle Rock Boss': [('bigkey', 'Turtle Rock')]}
    },
    'Turtle Rock Boss': {
        'type': 'dungeonboss', 'dungeon': 'Turtle Rock',
        'link': {
            'Turtle Rock Boss Item': [('and', [
                ('item', 'firerod'),
                ('item', 'icerod'),
                ('or', [
                    ('settings', 'placement_advanced'),
                    ('settings', 'swordless'),
                    ('settings', 'mastersword2'),
                    ('and', [
                        ('or', [
                            ('item', 'bottle'), ('item', 'halfmagic')]),
                        ('item', 'mastersword')])]),
                ('or', [
                    ('item', 'mastersword3'), ('item', 'hammer'),
                    ('and', [
                        ('or', [
                            ('item', 'bottle'), ('item', 'halfmagic')]),
                        ('item', 'sword')])])])]}
    },
    'Turtle Rock Boss Item': {
        'type': 'dungeonchest_nokey', 'dungeon': 'Turtle Rock',
        'link': {
            'Turtle Rock Reward': []}
    },
    'Turtle Rock Reward': {
        'type': 'dungeonreward', 'dungeon': 'Turtle Rock',
        'link': {
            'Turtle Rock Entrance Entrance (I)': []}
    },
}



'''
Single dungeon tracker
'''

import typing

from ..config.images import image

__all__ = 'DungeonObj',


class DungeonObj(object):
    '''
    Dungeon object

    Instance variables:
        identifier: name of dungeon
        location: (column, row) on dungeon display
        totalkeys: total number of small keys available for this dungeon
        icon: name of dungeon image
        features: additional features of dungeon
        tracker: world state tracker
        smallkeys: number of collected small keys
        bigkey: whether big key is collected or not
        chests: number of collected chests
        reward: dungeon reward
        completed: True if dungeon has been completed
        medallion: medallion requirement
    '''

    import_data = (
        'smallkeys', 'bigkey', 'chests', 'reward', 'completed', 'medallion')

    def __init__(self, name: str, location: (int, int), totalkeys: int,
                 icon: str, features: typing.Sequence[str], tracker):
        '''
        Args:
            name: name of dungeon
            location: (column, row) on dungeon display
            totalkeys: total number of randomly placed small keys
            icon: icon image of dungeon
            features: list of features of dungeon
            tracker: world state tracker
        '''

        self.identifier = name
        self.location = location
        self.totalkeys = totalkeys
        self.icon = image(icon)
        self.features = features
        self.tracker = tracker
        self.reset()

    def reset(self) -> None:
        '''
        Reset all values to default.
        '''

        self.smallkeys = 0
        self.bigkey = False
        self.tracker.keys[self.identifier] = {'small': 0, 'big': False}
        self.chests = 0
        self.reward = 'unknown'
        self.completed = False
        self.medallion = 'unknown'
        self.tracker.set_dungeon(self.identifier, self.export())

    def restore(self, data: dict) -> None:
        '''
        Rebuild dungeon info from save.

        Args:
            data: dungeon save data
        '''
        for attr in self.import_data:
            if attr in ('smallkeys', 'chests'):
                data[attr] = int(data[attr])
            self.set(attr, data[attr])
        self.tracker.keys[self.identifier]['small'] = data['smallkeys']
        self.tracker.keys[self.identifier]['big'] = data['bigkey']

    def save(self) -> dict:
        '''
        Return non-static dungeon info.

        Returns:
            dict: dungeon save data
        '''

        data = {}
        for attr in self.import_data:
            data[attr] = self.get(attr)
        return data

    def export(self) -> dict:
        '''
        Retrieve dungeon info for world state tracker.

        Returns:
            dict: dungeon info
        '''

        data = self.save()
        data['features'] = self.features
        return data

    def set(self, variable: str, value) -> None:
        '''
        Set dungeon variable.

        Args:
            variable: name of dungeon variable to change
            value: value to set variable to
        '''

        self.__setattr__(variable, value)
        self.tracker.set_dungeon(self.identifier, self.export())

    def get(self, variable: str) -> object:
        '''
        Get dungeon variable.

        Args:
            variable: name of dungeon variable
        Returns:
            object: current value of variable
        '''

        return self.__getattribute__(variable)

    def total_items(self) -> int:
        '''
        Retrieve number of items available in dungeon.

        Returns:
            int: number of non-dungeon-specific items
        '''

        if 'majoronly' in self.tracker.settings:
            return self.tracker.total_chests(self.identifier)
        total_chests = self.tracker.total_chests(self.identifier)
        if 'random_map' not in self.tracker.settings and 'map' in self.features:
            total_chests -= 1
        if ('random_compass' not in self.tracker.settings and
            'compass' in self.features):
            total_chests -= 1
        if ('random_bigkey' not in self.tracker.settings and
            'bigkey' in self.features):
            total_chests -= 1
        if ('random_smallkey' not in self.tracker.settings and
            'retro' not in self.tracker.settings):
            total_chests -= self.totalkeys
        return total_chests

    def remaining(self) -> int:
        '''
        Retrieve number of remaining items in dungeon.

        Returns:
            int: number of remaining items
        '''

        return self.total_items() - self.chests

    def cycle_reward(self, direction: bool) -> None:
        '''
        Cycle through dungeon reward.

        Args:
            direction: cycling direction
        '''

        circle = ('unknown', 'courage', 'powerwisdom', 'crystal', '56crystal')
        newidx = circle.index(self.reward)
        newidx += 1 if direction else -1
        if newidx >= len(circle):
            newidx = 0
        self.set('reward', circle[newidx])

    def mark_complete(self, complete: bool) -> None:
        '''
        Mark dungeon as (in)complete.

        Args:
            complete: False if dungeon is to be marked as completed
        '''

        self.set('completed', complete)
        self.tracker.set_dungeon_state(
            self.identifier, [bool(self.remaining()), not complete])

    def cycle_medallion(self, direction) -> None:
        '''
        Cycle through medallion requirement.

        Args:
            direction: cycling direction
        '''

        circle = ('unknown', 'bombos', 'ether', 'quake')
        newidx = circle.index(self.medallion)
        newidx += 1 if direction else -1
        if newidx >= len(circle):
            newidx = 0
        self.set('medallion', circle[newidx])

    def toggle_bigkey(self, *args) -> None:
        '''
        Toggle possession of the Big Key.
        '''

        self.tracker.keys[self.identifier]['big'] = not self.bigkey
        self.set('bigkey', not self.bigkey)

    def key_up(self, *args) -> None:
        '''
        Add one small key.
        '''

        before = self.get('smallkeys')
        after = min(self.smallkeys + 1, self.totalkeys)
        self.tracker.keys[self.identifier]['small'] += (after - before)
        self.set('smallkeys', after)

    def key_down(self, *args) -> None:
        '''
        Remove one small key.
        '''

        before = self.get('smallkeys')
        after = max(self.smallkeys - 1, 0)
        self.tracker.keys[self.identifier]['small'] += (after - before)
        self.set('smallkeys', after)

    def item_up(self, *args) -> None:
        '''
        Add one collected item.
        '''

        self.set('chests', min(self.chests + 1, self.total_items()))
        self.tracker.set_dungeon_state(
            self.identifier, [bool(self.remaining()), not self.completed])

    def item_down(self, *args) -> None:
        '''
        Remove one collected item.
        '''

        self.set('chests', max(self.chests - 1, 0))
        self.tracker.set_dungeon_state(
            self.identifier, [bool(self.remaining()), not self.completed])



'''
Dark Palace
'''

__all__ = 'LOCATIONS',


LOCATIONS = {
    'Dark Palace Entrance (E)': {
        'type': 'entrance_dungeon', 'map': 'dark', 'coord': (629, 262),
        'link': {
            'East Dark World': [],
            'Dark Palace Entrance (I)': [
                ('settings', 'placement_advanced'),
                ('and', [
                    ('or', [
                        ('settings', 'swordless'), ('item', 'sword')]),
                    ('item', 'bottle')])]}
    },
    'Dark Palace Entrance (I)': {
        'type': 'interior',
        'link': {
            'Dark Palace Entrance (E)': [],
            'Dark Palace Front': [('and', [
                ('rabbitbarrier', None),
                ('or', [
                    ('settings', 'placement_advanced'),
                    ('and', [
                        ('or', [
                            ('settings', 'swordless'), ('item', 'sword')]),
                        ('item', 'bottle')])])])]}
    },
    'Dark Palace Front': {
        'type': 'area', 'dungeon': 'Dark Palace',
        'link': {
            'Dark Palace Entrance (I)': [],
            'Dark Palace First Key': [],
            'Dark Palace Centre': [('smallkey', 'Dark Palace')],
            'Dark Palace Hidden Path': [
                ('item', 'bombs'), ('item', 'pegasus')]}
    },
    'Dark Palace First Key': {
        'type': 'dungeonchest', 'dungeon': 'Dark Palace',
        'link': {
            'Dark Palace Front': []}
    },
    'Dark Palace Hidden Path': {
        'type': 'area', 'dungeon': 'Dark Palace',
        'link': {
            'Dark Palace Front': [('item', 'bombs'), ('item', 'pegasus')],
            'Dark Palace Moving Floor': [
                ('item', 'bow'), ('settings', 'enemiser')]}
    },
    'Dark Palace Moving Floor': {
        'type': 'area', 'dungeon': 'Dark Palace',
        'link': {
            'Dark Palace Hidden Path': [],
            'Dark Palace Key With A View': [('item', 'bombs')],
            'Dark Palace Map': [],
            'Dark Palace Centre': [('item', 'hammer')]}
    },
    'Dark Palace Key With A View': {
        'type': 'dungeonchest', 'dungeon': 'Dark Palace',
        'link': {
            'Dark Palace Moving Floor': []}
    },
    'Dark Palace Map': {
        'type': 'dungeonchest', 'dungeon': 'Dark Palace',
        'link': {
            'Dark Palace Moving Floor': []}
    },
    'Dark Palace Centre': {
        'type': 'area', 'dungeon': 'Dark Palace',
        'link': {
            'Dark Palace Front': [],
            'Dark Palace Gauntlet Entrance': [('bigkey', 'Dark Palace')],
            'Dark Palace Big Key Drop': [('item', 'bombs')],
            'Dark Palace Stalfos Arena': [],
            'Dark Palace Targetting Chest': [],
            'Dark Palace Back': [('smallkey', 'Dark Palace')]}
    },
    'Dark Palace Big Key Drop': {
        'type': 'area', 'dungeon': 'Dark Palace',
        'link': {
            'Dark Palace Stalfos Arena': [],
            'Dark Palace Big Key': [('smallkey', 'Dark Palace')]}
    },
    'Dark Palace Big Key': {
        'type': 'dungeonchest', 'dungeon': 'Dark Palace',
        'link': {
            'Dark Palace Big Key Drop': [],
            'Dark Palace Big Key': []}
    },
    'Dark Palace Stalfos Arena': {
        'type': 'area', 'dungeon': 'Dark Palace',
        'link': {
            'Dark Palace Stalfos Key': [],
            'Dark Palace Front': []}
    },
    'Dark Palace Stalfos Key': {
        'type': 'dungeonchest', 'dungeon': 'Dark Palace',
        'link': {
            'Dark Palace Stalfos Arena': []}
    },
    'Dark Palace Targetting Chest': {
        'type': 'dungeonchest', 'dungeon': 'Dark Palace',
        'link': {
            'Dark Palace Centre': []}
    },
    'Dark Palace Back': {
        'type': 'area', 'dungeon': 'Dark Palace',
        'link': {
            'Dark Palace Maze Entrance': [('smallkey', 'Dark Palace')],
            'Dark Palace Treasury': [
                ('item', 'lantern'),
                ('and', [
                    ('settings', 'placement_advanded'), ('item', 'firerod')])],
            'Dark Palace Compass': [],
            'Dark Palace Return Path': []}
    },
    'Dark Palace Maze Entrance': {
        'type': 'area', 'dungeon': 'Dark Palace',
        'link': {
            'Dark Palace Back': [],
            'Dark Palace Maze': [('item', 'lantern')]}
    },
    'Dark Palace Maze': {
        'type': 'area', 'dungeon': 'Dark Palace',
        'link': {
            'Dark Palace Maze Entrance': [],
            'Dark Palace Maze Bombs': [],
            'Dark Palace Maze Key': [],
            'Dark Palace Treasure': [('item', 'bombs')]},
    },
    'Dark Palace Maze Bombs': {
        'type': 'dungeonchest', 'dungeon': 'Dark Palace',
        'link': {
            'Dark Palace Maze': []}
    },
    'Dark Palace Maze Key': {
        'type': 'dungeonchest', 'dungeon': 'Dark Palace',
        'link': {
            'Dark Palace Maze': []}
    },
    'Dark Palace Treasure': {
        'type': 'dungeonchest_nokey', 'dungeon': 'Dark Palace',
        'link': {
            'Dark Palace Maze': [('item', 'lantern')]}
    },
    'Dark Palace Treasury': {
        'type': 'area', 'dungeon': 'Dark Palace',
        'link': {
            'Dark Palace Back': [],
            'Dark Palace Treasury Left Chest': [],
            'Dark Palace Treasury Right Chest': []}
    },
    'Dark Palace Treasury Left Chest': {
        'type': 'dungeonchest', 'dungeon': 'Dark Palace',
        'link': {
            'Dark Palace Treasury': []}
    },
    'Dark Palace Treasury Right Chest': {
        'type': 'dungeonchest', 'dungeon': 'Dark Palace',
        'link': {
            'Dark Palace Treasury': []}
    },
    'Dark Palace Compass': {
        'type': 'dungeonchest', 'dungeon': 'Dark Palace',
        'link': {
            'Dark Palace Back': []}
    },
    'Dark Palace Return Path': {
        'type': 'area', 'dungeon': 'Dark Palace',
        'link': {
            'Dark Palace Back': [],
            'Dark Palace Return Chest': [],
            'Dark Palace Centre': []}
    },
    'Dark Palace Return Chest': {
        'type': 'dungeonchest', 'dungeon': 'Dark Palace',
        'link': {
            'Dark Palace Return Path': []}
    },
    'Dark Palace Gauntlet Entrance': {
        'type': 'area', 'dungeon': 'Dark Palace',
        'link': {
            'Dark Palace Centre': [],
            'Dark Palace Gauntlet Door': [('item', 'bow')]}
    },
    'Dark Palace Gauntlet Door': {
        'type': 'area', 'dungeon': 'Dark Palace',
        'link': {
            'Dark Palace Gauntlet Entrance': [],
            'Dark Palace Gauntlet': [('and', [
                ('item', 'lantern'), ('item', 'hammer')])]}
    },
    'Dark Palace Gauntlet': {
        'type': 'area', 'dungeon': 'Dark Palace',
        'link': {
            'Dark Palace Gauntlet Door': [('item', 'hammer')],
            'Dark Palace Boss': [('smallkey', 'Dark Palace')]}
    },
    'Dark Palace Boss': {
        'type': 'dungeonboss', 'dungeon': 'Dark Palace',
        'link': {
            'Dark Palace Boss Item': [('and', [
                ('or', [
                    ('item', 'bombs'), ('item', 'hammer')]),
                ('or', [
                    ('item', 'mastersword'), ('item', 'bow')])])]}
    },
    'Dark Palace Boss Item': {
        'type': 'dungeonchest_nokey', 'dungeon': 'Dark Palace',
        'link': {
            'Dark Palace Reward': []}
    },
    'Dark Palace Reward': {
        'type': 'dungeonreward', 'dungeon': 'Dark Palace',
        'link': {
            'Dark Palace Entrance (I)': []}
    },
}



'''
Dark South World
'''

__all__ = 'LOCATIONS',


LOCATIONS = {
    'South Dark World': {
        'type': 'area',
        'link': {
            'South Light World': [('and', [
                ('nosettings', 'inverted'), ('item', 'mirror')])],
            'Southern Shores': [('and', [
                ('nosettings', 'inverted'), ('item', 'mirror')])],
            'Running Game': [('and', [
                ('nosettings', 'inverted'), ('item', 'mirror')])],
            'Cave Near Haunted Grove Entrance (E)': [('and', [
                ('nosettings', 'inverted'), ('item', 'mirror')])],
            'Bombos Stone Mountain': [('and', [
                ('nosettings', 'inverted'), ('item', 'mirror')])],
            'Dark Bottle Ocarina': [],
            'Digging Game': [('rabbitbarrier', None)],
            'Shooting Game Entrance (E)': [],
            "Thieves' Town Surface": [('item', 'titansmitts')],
            'Ocarina Boy': [('rabbitbarrier', None)],
            'Fairy Pond Under Rocks (DW) Entrance (E)': [('item', 'pegasus')],
            'Portal Ersatz': [('and', [
                ('settings', 'inverted'), ('access', 'Castle Tower Reward'),
                ('rabbitbarrier', None)])],
            'Bomb Shop Entrance (E)': [],
            'East Dark World': [('item', 'hammer')],
            'Frozen Lake Shop Entrance (E)': [],
            'Frozen Lake': [('item', 'flippers')],
            'South Portal (DW)': [('item', 'hammer')],
            'Hiding Thief Cave Entrance (E)': [('item', 'bombs')],
            'Swamp Palace Entrance (E)': [],
            'Misery Mire Surface': [('and', [
                ('glitch', 'overworld'), ('item', 'pegasus')])]}
    },
    'Digging Game': {
        'type': 'item', 'map': 'dark', 'coord': (32, 459),
        'link': {
            'South Dark World': []}
    },
    'Shooting Game Entrance (E)': {
        'type': 'entrance', 'map': 'dark', 'coord': (137, 464),
        'link': {
            'South Dark World': [],
            'Shooting Game Entrance (I)': []}
    },
    'Shooting Game Entrance (I)': {
        'type': 'interior',
        'link': {
            'Shooting Game Entrance (E)': []}
    },
    'Ocarina Boy': {
        'type': 'item', 'map': 'dark', 'coord': (199, 452),
        'link': {
            'South Dark World': []}
    },
    'Fairy Pond Under Rocks (DW) Entrance (E)': {
        'type': 'entrance', 'map': 'dark', 'coord': (307, 433),
        'link': {
            'South Dark World': [],
            'Fairy Pond Under Rocks (DW) Entrance (I)': []}
    },
    'Fairy Pond Under Rocks (DW) Entrance (I)': {
        'type': 'interior',
        'link': {
            'Fairy Pond Under Rocks (DW) Entrance (E)': []}
    },
    'Portal Ersatz': {
        'type': 'area',
        'link': {
            'South Light World': [('state', 'rabbit')]}
    },
    'Bomb Shop Entrance (E)': {
        'type': 'entrance_unique', 'map': 'dark', 'coord': (363, 456),
        'link': {
            'South Dark World': [],
            'Bomb Shop Entrance (I)': [('nosettings', 'inverted')],
            "Link's House Entrance (I)": [('settings', 'inverted')]}
    },
    'Bomb Shop Entrance (I)': {
        'type': 'interior',
        'link': {
            'Bomb Shop Entrance (E)': [('nosettings', 'inverted')],
            "Link's House Entrance (E)": [('settings', 'inverted')]}
    },
    'Frozen Lake Shop Entrance (E)': {
        'type': 'entrance_shop', 'map': 'dark', 'coord': (424, 532),
        'link': {
            'South Dark World': [],
            'Frozen Lake Shop Entrance (I)': []}
    },
    'Frozen Lake Shop Entrance (I)': {
        'type': 'interior',
        'link': {
            'Frozen Lake Shop Entrance (E)': []}
    },
    'South Portal (DW)': {
        'type': 'area',
        'link': {
            'South Portal (LW)': [('and', [
                ('nosettings', 'inverted'), ('item', 'mirror')])],
            'South Dark World': [('item', 'hammer')],
            'South Portal': [('item', 'powerglove')]}
    },
    'Hiding Thief Cave Entrance (E)': {
        'type': 'entrance_unique', 'map': 'dark', 'coord': (391, 515),
        'link': {
            'South Dark World': [],
            'Hiding Thief Cave Entrance (I)': []}
    },
    'Hiding Thief Cave Entrance (I)': {
        'type': 'interior',
        'link': {
            'Hiding Thief Cave Entrance (E)': [],
            'Hiding Thief Cave': []}
    },
    'Hiding Thief Cave': {
        'type': 'cave', 'map': 'dark', 'coord': (391, 515),
        'link': {
            'Hiding Thief Cave Entrance (I)': []}
    },

    'Frozen Lake': {
        'type': 'area',
        'link': {
            'Lake Hylia': [('and', [
                ('nosettings', 'inverted'), ('item', 'mirror')])],
            'Small Lake Hylia Island': [('and', [
                ('nosettings', 'inverted'), ('item', 'mirror')])],
            'Ice Palace Exterior': [('settings', 'inverted')],
            'East Dark World': [],
            'Dark World Waterway': [],
            'Eastern Shores': []}
    },
    'Ice Palace Exterior': {
        'type': 'area',
        'link': {
            'Large Lake Hylia Island': [('and', [
                ('nosettings', 'inverted'), ('item', 'mirror')])],
            'Ice Palace Portal': [('item', 'titansmitts')],
            'Frozen Lake': [('and', [
                ('settings', 'inverted'), ('item', 'flippers')])],
            'Ice Palace Entrance (E)': []}
    },
    'Dark World Waterway': {
        'type': 'area',
        'link': {
            'Frozen Lake': [],
            'Dark River': []}
    },

    'Eastern Shores': {
        'type': 'area',
        'link': {
            'Southern Shores': [('and', [
                ('nosettings', 'inverted'), ('item', 'mirror')])],
            'Frozen Lake': [('item', 'flippers')],
            'Eastern Shores Under Rock Entrance (E)': [('item', 'powerglove')],
            'Eastern Shores Fairy Entrance (E)': [('item', 'bombs')],
            'Eastern Shores Cave Entrance (E)': []}
    },
    'Eastern Shores Under Rock Entrance (E)': {
        'type': 'entrance', 'map': 'dark', 'coord': (592, 521),
        'link': {
            'Eastern Shores': [],
            'Eastern Shores Under Rock Entrance (I)': []}
    },
    'Eastern Shores Under Rock Entrance (I)': {
        'type': 'interior',
        'link': {
            'Eastern Shores Under Rock Entrance (I)': []}
    },
    'Eastern Shores Fairy Entrance (E)': {
        'type': 'entrance', 'map': 'dark', 'coord': (588, 507),
        'link': {
            'Eastern Shores': [],
            'Eastern Shores Fairy Entrance (I)': []}
    },
    'Eastern Shores Fairy Entrance (I)': {
        'type': 'interior',
        'link': {
            'Eastern Shores Fairy Entrance (E)': []}
    },
    'Eastern Shores Cave Entrance (E)': {
        'type': 'entrance', 'map': 'dark', 'coord': (600, 507),
        'link': {
            'Eastern Shores': [],
            'Eastern Shores Cave Entrance (I)': []}
    },
    'Eastern Shores Cave Entrance (I)': {
        'type': 'interior',
        'link': {
            'Eastern Shores Cave Entrance (E)': []}
    },
}



'''
Update checker
'''

import http.client
import json

from .. import version

__all__ = 'check_update_availability',


def check_update_availability() -> str:
    '''
    Check whether update is available.

    Returns:
        str: version string if update is available, else False
    '''

    print('Current version is {0:s}.'.format(version.__version__))
    current = tuple(int(v) for v in version.__version__.split('.'))

    print('Checking for updates ...')
    request = http.client.HTTPSConnection('pypi.org')
    request.request('GET', '/pypi/z3-tracker/json')
    try:
        ret = request.getresponse()
    except http.client.HTTPException as err:
        print('   ... failed. Got: {0:s}'.format(str(err)))
        return False
    if ret.status != 200:
        print('   ... failed. Got: {0:d}-{1:s}'.format(
            ret.status, ret.reason))
        return False
    ret = ret.read()
    try:
        ret = ret.decode()
    except UnicodeError as err:
        print('   ... failed. Got: {0:s}'.format(str(err)))
        return False
    try:
        ret = json.loads(ret)
    except json.JSONDecoder as err:
        print('   ... failed. Got: {0:s}'.format(str(err)))
        return False
    try:
        newest = ret['info']['version']
    except KeyError:
        print('   ... failed. Could not find any version information.')
        return False
    print('   ... done.')

    updated = tuple(int(v) for v in newest.split('.'))
    for m in range(min(len(current), len(updated))):
        if updated[m] > current[m]:
            update_available = True
            break
    else:
        if len(updated) > len(current):
            update_available = True
        else:
            update_available = False

    if update_available:
        print('Found newer version: {0:s}'.format(newest))
        return newest
    print('No newer version found.')
    return False



'''
Desert
'''

__all__ = 'LOCATIONS',


LOCATIONS = {
    'Desert': {
        'type': 'area',
        'link': {
            'Misery Mire Surface': [('and', [
                ('settings', 'inverted'), ('item', 'mirror')])],
            'Light Bottle Ocarina': [],
            'Southern Shores': [],
            "Aginah's Cave Entrance (E)": [],
            'Desert Cave Valley': [
                ('settings', 'inverted'),
                ('glitch', 'major'),
                ('and', [
                    ('glitch', 'overworld'), ('item', 'pegasus')])],
            'Desert Palace Stairs': [('item', 'mudora')],
            'Desert Portal (LW)': [('and', [
                ('glitch', 'overworld'), ('item', 'pegasus')])],
            'Desert Ridge': [('and', [
                ('glitch', 'overworld'), ('item', 'pegasus')])]},
        'visible': {"Vulture's Heart": []}
    },
    "Aginah's Cave Entrance (E)": {
        'type': 'entrance_unique', 'map': 'light', 'coord': (132, 546),
        'link': {
            'Desert': [],
            "Aginah's Cave Entrance (I)": []}
    },
    "Aginah's Cave Entrance (I)": {
        'type': 'interior',
        'link': {
            "Aginah's Cave Entrance (E)": [],
            "Aginah's Cave Interior": []}
    },
    "Aginah's Cave Interior": {
        'type': 'area',
        'link': {
            "Aginah's Cave Entrance (I)": [],
            "Aginah's Cave": [('item', 'bombs')]}
    },
    "Aginah's Cave": {
        'type': 'cave', 'map': 'light', 'coord': (132, 546),
        'link': {
            "Aginah's Cave Interior": []}
    },
    'Desert Cave Valley': {
        'type': 'area',
        'link': {
            'Misery Mire Surface': [('and', [
                ('settings', 'inverted'), ('item', 'mirror')])],
            'Desert': [],
            'Desert Cave Entrance (E)': [('item', 'powerglove')]}
    },
    'Desert Cave Entrance (E)': {
        'type': 'entrance_unique', 'map': 'light', 'coord': (116, 515),
        'link': {
            'Desert Cave Valley': [],
            'Desert Cave Entrance (I)': []}
    },
    'Desert Cave Entrance (I)': {
        'type': 'interior',
        'link': {
            'Desert Cave Entrance (E)': [],
            'Desert Cave': []}
    },
    'Desert Cave': {
        'type': 'drop', 'map': 'light', 'coord': (116, 515),
        'link': {
            'Desert Cave Entrance (I)': []}
    },

    'Desert Portal (LW)': {
        'type': 'area',
        'link': {
            'Desert Portal (DW)': [('and', [
                ('settings', 'inverted'), ('item', 'mirror')])],
            'Desert Portal': [('item', 'titansmitts')],
            'Desert': []}
    },
    'Desert Portal': {
        'type': 'area',
        'link': {
            'Desert Portal (DW)': [('and', [
                ('nosettings', 'inverted'), ('state', 'rabbit')])],
            'Desert Portal (LW)': [('and', [
                ('settings', 'inverted'), ('state', 'rabbit')])]}
    },

    'Desert Palace Stairs': {
        'type': 'area',
        'link': {
            'Misery Mire Surface': [('and', [
                ('settings', 'inverted'), ('item', 'mirror')])],
            'Desert Palace Main Entrance Entrance (E)': []}
    },
    'Desert Palace Main Entrance Entrance (E)': {
        'type': 'entrance_dungeon', 'map': 'light', 'coord': (49, 549),
        'link': {
            'Desert Palace Stairs': [],
            'Desert Palace Main Entrance Entrance (I)': []}
    },
    'Desert Palace Main Entrance Entrance (I)': {
        'type': 'interior',
        'link': {
            'Desert Palace Main Entrance Entrance (E)': [],
            'Desert Palace Interior': [('and', [
                ('rabbitbarrier', None),
                ('or', [('nosettings', 'majoronly'), ('item', 'pegasus')])])]}
    },
    'Desert Palace Side Entrance Entrance (I)': {
        'type': 'interior',
        'link': {
            'Desert Palace Side Entrance Entrance (E)': [],
            'Desert Palace Interior': [('and', [
                ('rabbitbarrier', None),
                ('or', [('nosettings', 'majoronly'), ('item', 'pegasus')])])]}
    },
    'Desert Palace Side Entrance Entrance (E)': {
        'type': 'entrance_dungeon', 'map': 'light', 'coord': (75, 527),
        'link': {
            'Desert': [],
            'Desert Palace Side Entrance Entrance (I)': []}
    },

    'Desert Palace Interior': {
        'type': 'area', 'dungeon': 'Desert Palace',
        'link': {
            'Desert Palace Main Entrance Entrance (I)': [],
            'Desert Palace Side Entrance Entrance (I)': [],
            'Desert Palace Back Entrance Entrance (I)': [],
            'Desert Palace Compass': [('smallkey', 'Desert Palace')],
            'Desert Palace Map': [],
            'Desert Palace Small Key': [('item', 'pegasus')],
            'Desert Palace Treasure': [('bigkey', 'Desert Palace')]}
    },
    'Desert Palace Compass': {
        'type': 'dungeonchest', 'dungeon': 'Desert Palace',
        'link': {
            'Desert Palace Interior': [],
            'Desert Palace Big Key': []}
    },
    'Desert Palace Big Key': {
        'type': 'dungeonchest', 'dungeon': 'Desert Palace',
        'link': {
            'Desert Palace Compass': []}
    },
    'Desert Palace Map': {
        'type': 'dungeonchest', 'dungeon': 'Desert Palace',
        'link': {
            'Desert Palace Interior': []}
    },
    'Desert Palace Small Key': {
        'type': 'dungeonchest', 'dungeon': 'Desert Palace',
        'link': {
            'Desert Palace Interior': []}
    },
    'Desert Palace Treasure': {
        'type': 'dungeonchest', 'dungeon': 'Desert Palace',
        'link': {
            'Desert Palace Interior': []}
    },
    'Desert Palace Back Entrance Entrance (I)': {
        'type': 'interior',
        'link': {
            'Desert Palace Back Entrance Entrance (E)': [],
            'Desert Palace Interior': [('and', [
                ('rabbitbarrier', None),
                ('or', [
                    ('nosettings', 'majoronly'), ('item', 'pegasus')])])]}
    },
    'Desert Palace Back Entrance Entrance (E)': {
        'type': 'entrance_dungeon', 'map': 'light', 'coord': (23, 527),
        'link': {
            'Desert Ridge': [],
            'Desert Palace Back Entrance Entrance (I)': []}
    },

    'Desert Ridge': {
        'type': 'area',
        'link': {
            'Misery Mire Surface': [('and', [
                ('settings',  'inverted'), ('item', 'mirror')])],
            'Desert Palace Back Entrance Entrance (E)': [],
            'Desert': [],
            "Vulture's Heart": [],
            'Desert Palace Gauntlet Entrance': [('item', 'powerglove')]}
    },
    "Vulture's Heart": {
        'type': 'item', 'map': 'light', 'coord': (16, 606),
        'link': {
            'Desert Ridge': []}
    },
    'Desert Palace Gauntlet Entrance': {
        'type': 'area',
        'link': {
            'Misery Mire Surface': [('and', [
                ('settings', 'inverted'), ('item', 'mirror')])],
            'Desert Ridge': [('item', 'powerglove')],
            'Desert Palace Gauntlet Entrance (E)': []}
    },
    'Desert Palace Gauntlet Entrance (E)': {
        'type': 'entrance_dungeon', 'map': 'light', 'coord': (49, 507),
        'link': {
            'Desert Palace Gauntlet Entrance': [],
            'Desert Palace Gauntlet Entrance (I)': []}
    },
    'Desert Palace Gauntlet Entrance (I)': {
        'type': 'interior',
        'link': {
            'Desert Palace Gauntlet Entrance (E)': [],
            'Desert Palace Trap Room 1': [('rabbitbarrier', None)]}
    },
    'Desert Palace Trap Room 1': {
        'type': 'area', 'dungeon': 'Desert Palace',
        'link': {
            'Desert Palace Gauntlet Entrance (I)': [],
            'Desert Palace Hidden Key 1': [],
            'Desert Palace Trap Room 2': [('smallkey', 'Desert Palace')]}
    },
    'Desert Palace Hidden Key 1': {
        'type': 'dungeonkey', 'dungeon': 'Desert Palace',
        'link': {
            'Desert Palace Trap Room 1': []}
    },
    'Desert Palace Trap Room 2': {
        'type': 'area', 'dungeon': 'Desert Palace',
        'link': {
            'Desert Palace Trap Room 1': [],
            'Desert Palace Hidden Key 2': [],
            'Desert Palace Torch Room': [('smallkey', 'Desert Palace')]}
    },
    'Desert Palace Hidden Key 2': {
        'type': 'dungeonkey', 'dungeon': 'Desert Palace',
        'link': {
            'Desert Palace Trap Room 2': []}
    },
    'Desert Palace Torch Room': {
        'type': 'area', 'dungeon': 'Desert Palace',
        'link': {
            'Desert Palace Trap Room 2': [],
            'Desert Palace Boss Door': [
                ('item', 'lantern'), ('item', 'firerod')]}
    },
    'Desert Palace Boss Door': {
        'type': 'area', 'dungeon': 'Desert Palace',
        'link': {
            'Desert Palace Torch Room': [],
            'Desert Palace Boss': [('bigkey', 'Desert Palace')]}
    },
    'Desert Palace Boss': {
        'type': 'dungeonboss', 'dungeon': 'Desert Palace',
        'link': {
            'Desert Palace Boss Item': [
                ('item', 'sword'), ('item', 'hammer'), ('item', 'bow'),
                ('item', 'firerod'), ('item', 'icerod'), ('item', 'byrna'),
                ('item', 'somaria')]}
        },
    'Desert Palace Boss Item': {
        'type': 'dungeonchest_nokey', 'dungeon': 'Desert Palace',
        'link': {
            'Desert Palace Reward': []}
    },
    'Desert Palace Reward': {
        'type': 'dungeonreward', 'dungeon': 'Desert Palace',
        'link': {
            'Desert Palace Main Entrance Entrance (I)': []}
    },
}



'''
Mount Hebra
'''

__all__ = 'LOCATIONS',


LOCATIONS = {
    'Hebra Ascent': {
        'type': 'area',
        'link': {
            'Hebra Ascent (Bottom) Entrance (I)': [('item', 'lantern')],
            'Finding The Lost Man': [('item', 'lantern')],
            'Hebra Ascent (Top) Entrance (I)': [('item', 'lantern')]}
    },
    'Finding The Lost Man': {
        'type': 'area',
        'link': {
            'Hebra Ascent': []}
    },
    'Hebra Ascent (Top) Entrance (I)': {
        'type': 'interior',
        'link': {
            'Hebra Ascent': [],
            'Hebra Ascent (Top) Entrance (E)': [('nosettings', 'inverted')],
            'West Death Mountain Fairy Entrance (E)': [
                ('settings', 'inverted')]}
    },
    'Hebra Ascent (Top) Entrance (E)': {
        'type': 'entrance_unique', 'map': 'light', 'coord': (269, 125),
        'link': {
            'Hebra Ascent (Top) Entrance (I)': [('nosettings', 'inverted')],
            'West Death Mountain Fairy Entrance (I)': [
                ('settings', 'inverted')],
            'West Lower Mount Hebra': []}
    },
    'West Lower Mount Hebra': {
        'type': 'area',
        'link': {
            'West Lower Death Mountain': [('and', [
                ('settings', 'inverted'), ('item', 'mirror')])],
            'Light Bottle Ocarina': [],
            'Hebra Ascent (Top) Entrance (E)': [],
            "Lost Man's Cave Front Entrance Entrance (E)": [],
            "Lost Man's Cave Back Entrance Entrance (E)": [],
            'East Lower Mount Hebra': [
                ('item', 'hookshot'),
                ('glitch', 'major'),
                ('and', [
                    ('glitch', 'overworld'),
                    ('or', [
                        ('item', 'mirror'), ('item', 'pegasus')])])],
            'Hebra Descent (Top) Entrance (E)': [],
            'Heart Rock Cave Top Entrance Entrance (E)': [],
            'West Hebra Fairy Cave Entrance (E)': [],
            'Heart Rock Cave Bottom Entrance Entrance (E)': [],
            'Hebra Portal': [],
            'Pyramid Fairy Entrance (E)': [('and', [
                ('glitch', 'overworld'), ('item', 'mirror'),
                ('item', 'pegasus'),
                ('or', [
                    ('item', 'sword'), ('item', 'hookshot')])])],
            'Heart Rock': [
                ('and', [
                    ('glitch', 'overworld'), ('item', 'pegasus')]),
                ('glitch', 'major')],
            "King's Tomb Entrance (E)": [('and', [
                ('glitch', 'overworld'), ('item', 'pegasus')])]},
        'visible': {'Heart Rock': []}
    },
    "Lost Man's Cave Front Entrance Entrance (E)": {
        'type': 'entrance_unique', 'map': 'light', 'coord': (298, 155),
        'link': {
            'West Lower Mount Hebra': [],
            "Lost Man's Cave Front Entrance Entrance (I)": []}
    },
    "Lost Man's Cave Front Entrance Entrance (I)": {
        'type': 'interior',
        'link': {
            "Lost Man's Cave Front Entrance Entrance (E)": [],
            "Lost Man's Cave Front": []}
    },
    "Lost Man's Cave Front": {
        'type': 'area',
        'link': {
            "Lost Man's Cave Front Entrance Entrance (I)": [],
            'Lost Man': [('access', 'Finding The Lost Man')],
            "Lost Man's Cave Back": [('item', 'lantern')]}
    },
    'Lost Man': {
        'type': 'chest', 'map': 'light', 'coord': (298, 150),
        'link': {
            "Lost Man's Cave Front": []}
    },
    "Lost Man's Cave Back": {
        'type': 'area',
        'link': {
            "Lost Man's Cave Front": [],
            "Lost Man's Cave Back Entrance Entrance (I)": []}
    },
    "Lost Man's Cave Back Entrance Entrance (I)": {
        'type': 'interior',
        'link': {
            "Lost Man's Cave Front": [('item', 'lantern')],
            "Lost Man's Cave Back Entrance Entrance (E)": []}
    },
    "Lost Man's Cave Back Entrance Entrance (E)": {
        'type': 'entrance_unique', 'map': 'light', 'coord': (355, 106),
        'link': {
            "Lost Man's Cave Back Entrance Entrance (I)": [],
            'West Lower Mount Hebra': []}
    },
    'Hebra Descent (Top) Entrance (E)': {
        'type': 'entrance_unique', 'map': 'light', 'coord': (262, 91),
        'link': {
            'West Lower Mount Hebra': [],
            'Hebra Descent (Top) Entrance (I)': []}
    },
    'Hebra Descent (Top) Entrance (I)': {
        'type': 'interior',
        'link': {
            'Hebra Descent (Top) Entrance (E)': [],
            'Hebra Descent': []}
    },
    'Hebra Descent': {
        'type': 'area',
        'link': {
            'Hebra Descent (Bottom) Entrance (I)': [],
            'Hebra Descent (Top) Entrance (I)': []}
    },
    'Heart Rock Cave Top Entrance Entrance (E)': {
        'type': 'entrance_unique', 'map': 'light', 'coord': (324, 68),
        'link': {
            'West Lower Mount Hebra': [],
            'Heart Rock Cave Top Entrance Entrance (I)': []}
    },
    'Heart Rock Cave Top Entrance Entrance (I)': {
        'type': 'interior',
        'link': {
            'Heart Rock Cave Top Entrance Entrance (E)': [],
            'Heart Rock Cave Top': []}
    },
    'Heart Rock Cave Top': {
        'type': 'area',
        'link': {
            'Heart Rock Cave Top Entrance Entrance (I)': [],
            'West Hebra Fairy Cave': []}
    },
    'West Hebra Fairy Cave Entrance (E)': {
        'type': 'entrance_unique', 'map': 'light', 'coord': (302, 91),
        'link': {
            'West Lower Mount Hebra': [],
            'West Hebra Fairy Cave Entrance (I)': []}
    },
    'West Hebra Fairy Cave Entrance (I)': {
        'type': 'interior',
        'link': {
            'West Hebra Fairy Cave Entrance (E)': [],
            'West Hebra Fairy Cave': []}
    },
    'West Hebra Fairy Cave': {
        'type': 'area',
        'link': {
            'West Hebra Fairy Cave Entrance (I)': []}
    },
    'Heart Rock Cave Bottom Entrance Entrance (E)': {
        'type': 'entrance_unique', 'map': 'light', 'coord': (324, 96),
        'link': {
            'West Lower Mount Hebra': [],
            'Heart Rock Cave Bottom Entrance Entrance (I)': []}
    },
    'Heart Rock Cave Bottom Entrance Entrance (I)': {
        'type': 'interior',
        'link': {
            'Heart Rock Cave Bottom Entrance Entrance (E)': [],
            'Heart Rock Cave Bottom': []}
    },
    'Heart Rock Cave Bottom': {
        'type': 'area',
        'link': {
            'Heart Rock Cave Bottom Entrance Entrance (E)': [],
            'Heart Rock Cave Heart': [],
            'West Hebra Fairy Cave': []}
    },
    'Heart Rock Cave Heart': {
        'type': 'cave', 'map': 'light', 'coord': (324, 88),
        'link': {
            'Heart Rock Cave Bottom': []}
    },
    'Hebra Portal': {
        'type': 'area',
        'link': {
            'West Lower Death Mountain': [('and', [
                ('nosettings', 'inverted'), ('state', 'rabbit')])],
            'West Lower Mount Hebra': [('and', [
                ('settings', 'inverted'), ('state', 'rabbit')])]}
    },

    'Heart Rock': {
        'type': 'item', 'map': 'light', 'coord': (337, 55),
        'link': {
            'West Lower Death Mountain': [('and', [
                ('settings', 'inverted'), ('item', 'mirror')])],
            'West Lower Mount Hebra': [],
            'West Upper Mount Hebra': []}
    },
    'West Upper Mount Hebra': {
        'type': 'area',
        'link': {
            'West Upper Death Mountain': [('and', [
                ('settings', 'inverted'), ('item', 'mirror')])],
            'West Lower Mount Hebra': [],
            'Heart Rock': [('settings', 'inverted')],
            'Ether Rock': [('mudora', 'take')],
            'Mountain Tower Entrance (E)': [],
            'East Upper Mount Hebra': [('item', 'hammer')]},
        'visible': {'Heart Rock': [], 'Ether Rock': [('item', 'mudora')]}
    },
    'Ether Rock': {
        'type': 'item', 'map': 'light', 'coord': (278, 13),
        'link': {
            'West Upper Mount Hebra': []}
    },

    'East Lower Mount Hebra': {
        'type': 'area',
        'link': {
            'West Lower Death Mountain': [('and', [
                ('settings', 'inverted'), ('item', 'mirror')])],
            'West Lower Mount Hebra': [('item', 'hookshot')],
            'Spiral Cave Bottom Entrance Entrance (E)': [],
            'Fairy Maze Bottom Entrance Entrance (E)': [
                ('item', 'titansmitts')],
            'East Hebra Fairy Cave Entrance Entrance (E)': [],
            'Hebra Shop Cave Entrance (E)': [],
            'East Hebra Portal': [('item', 'titansmitts')],
            'East Hebra Cave Bottom Entrance Entrance (E)': []}
    },
    'Spiral Cave Bottom Entrance Entrance (E)': {
        'type': 'entrance_unique', 'map': 'light', 'coord': (530, 86),
        'link': {
            'East Lower Mount Hebra': [],
            'Spiral Cave Bottom Entrance Entrance (I)': []}
    },
    'Spiral Cave Bottom Entrance Entrance (I)': {
        'type': 'interior',
        'link': {
            'Spiral Cave Bottom Entrance Entrance (E)': [],
            'Spiral Cave Bottom': []}
    },
    'Spiral Cave Bottom': {
        'type': 'area',
        'link': {
            'Spiral Cave Bottom Entrance Entrance (I)': []}
    },
    'Fairy Maze Bottom Entrance Entrance (E)': {
        'type': 'entrance_unique', 'map': 'light', 'coord': (544, 91),
        'link': {
            'East Lower Mount Hebra': [],
            'Fairy Maze Bottom Entrance Entrance (I)': []}
    },
    'Fairy Maze Bottom Entrance Entrance (I)': {
        'type': 'interior',
        'link': {
            'Fairy Maze Bottom Entrance Entrance (E)': [],
            'Fairy Maze Top Entrance Entrance (I)': []}
    },
    'East Hebra Fairy Cave Entrance Entrance (E)': {
        'type': 'entrance', 'map': 'light', 'coord': (558, 96),
        'link': {
            'East Lower Mount Hebra': [],
            'East Hebra Fairy Cave Entrance Entrance (I)': []}
    },
    'East Hebra Fairy Cave Entrance Entrance (I)': {
        'type': 'interior',
        'link': {
            'East Hebra Fairy Cave Entrance Entrance (E)': []}
    },
    'Hebra Shop Cave Entrance (E)': {
        'type': 'entrance_shop', 'map': 'light', 'coord': (572, 96),
        'link': {
            'East Lower Mount Hebra': [],
            'Hebra Shop Cave Entrance (I)': []}
    },
    'Hebra Shop Cave Entrance (I)': {
        'type': 'interior',
        'link': {
            'Hebra Shop Cave Entrance (E)': [],
            'Hebra Shop Cave': []}
    },
    'Hebra Shop Cave': {
        'type': 'area',
        'link': {
            'Hebra Shop Cave Entrance (I)': []}
    },
    'East Hebra Portal': {
        'type': 'area',
        'link': {
            'East Lower Death Mountain': [('and', [
                ('nosettings', 'inverted'), ('state', 'rabbit')])],
            'East Lower Mount Hebra': [('and', [
                ('settings', 'inverted'), ('state', 'rabbit')])]}
    },
    'East Hebra Cave Bottom Entrance Entrance (E)': {
        'type': 'entrance_unique', 'map': 'light', 'coord': (573, 142),
        'link': {
            'East Lower Mount Hebra': [],
            'East Hebra Cave Bottom Entrance Entrance (I)': []}
    },
    'East Hebra Cave Bottom Entrance Entrance (I)': {
        'type': 'interior',
        'link': {
            'East Hebra Cave Bottom Entrance Entrance (E)': [],
            'East Hebra Cave (area)': []}
    },
    'East Hebra Cave (area)': {
        'type': 'area',
        'link': {
            'East Hebra Cave Bottom Entrance Entrance (I)': [],
            'Hebra Shop Cave': [('rabbitbarrier', None)],
            'East Hebra Cave': [('item', 'bombs')],
            'East Hebra Cave Top Entrance Entrance (I)': []}
    },
    'East Hebra Cave': {
        'type': 'cave', 'map': 'light', 'coord': (573, 142),
        'link': {
            'Hebra Shop Cave': []}
    },
    'East Hebra Cave Top Entrance Entrance (I)': {
        'type': 'interior',
        'link': {
            'East Hebra Cave (area)': [],
            'East Hebra Cave Top Entrance Entrance (E)': []}
    },
    'East Hebra Cave Top Entrance Entrance (E)': {
        'type': 'entrance_unique', 'map': 'light', 'coord': (569, 41),
        'link': {
            'East Hebra Cave (area)': [],
            'East Upper Mount Hebra': []}
    },

    'East Upper Mount Hebra': {
        'type': 'area',
        'link': {
            'East Upper Death Mountain': [('and', [
                ('settings', 'inverted'), ('item', 'mirror')])],
            'Turtle Rock Balcony': [('and', [
                ('settings', 'inverted'), ('item', 'mirror')])],
            'Turtle Rock Fairy Exit': [('and', [
                ('settings', 'inverted'), ('item', 'mirror')])],
            'Light Bottle Ocarina': [],
            'West Upper Mount Hebra': [('item', 'hammer')],
            'East Lower Mount Hebra': [],
            'Spiral Cave Top Entrance Entrance (E)': [],
            'Fairy Maze Top Entrance Entrance (E)': [],
            'Goriya Cave Entrance (E)': [('settings', 'inverted')],
            'East Hebra Cave Top Entrance Entrance (E)': [],
            'Heart Peak': [
                ('settings', 'inverted'),
                ('and', [
                    ('glitch', 'overworld'), ('item', 'mirror')])],
            'Turtle Rock Portal (LW)': [('item', 'titansmitts')]},
        'visible': {'Heart Peak': []}
    },
    'Spiral Cave Top Entrance Entrance (E)': {
        'type': 'entrance_unique', 'map': 'light', 'coord': (529, 59),
        'link': {
            'East Lower Mount Hebra': [],
            'Spiral Cave Top Entrance Entrance (I)': []}
    },
    'Spiral Cave Top Entrance Entrance (I)': {
        'type': 'interior',
        'link': {
            'Spiral Cave Top Entrance Entrance (E)': [],
            'Spiral Cave Top': []}
    },
    'Spiral Cave Top': {
        'type': 'area',
        'link': {
            'Spiral Cave Top Entrance Entrance (I)': [],
            'Spiral Cave': [('rabbitbarrier', None)]}
    },
    'Spiral Cave': {
        'type': 'cave', 'map': 'light', 'coord': (529, 59),
        'link': {
            'Spiral Cave Bottom': []}
    },
    'Fairy Maze Top Entrance Entrance (E)': {
        'type': 'entrance_unique', 'map': 'light', 'coord': (544, 75),
        'link': {
            'East Lower Mount Hebra': [],
            'Fairy Maze Top Entrance Entrance (I)': [],
            'Fairy Maze Bottom Entrance Entrance (E)': []}
    },
    'Fairy Maze Top Entrance Entrance (I)': {
        'type': 'interior',
        'link': {
            'Fairy Maze Top Entrance Entrance (E)': [],
            'Fairy Maze Bottom Entrance Entrance (I)': []}
    },
    'Goriya Cave Entrance (E)': {
        'type': 'entrance_unique', 'map': 'light', 'coord': (559, 59),
        'link': {
            'East Lower Mount Hebra': [],
            'Goriya Cave Entrance (I)': []}
    },
    'Goriya Cave Entrance (I)': {
        'type': 'interior',
        'link': {
            'Goriya Cave Entrance (E)': [],
            'Goriya Cave': [('item', 'hammer')]}
    },
    'Goriya Cave': {
        'type': 'cave', 'map': 'light', 'coord': (559, 59),
        'link': {
            'Goriya Cave Entrance (I)': [('item', 'hammer')]}
    },
    'Heart Peak': {
        'type': 'item', 'map': 'light', 'coord': (539, 14),
        'link': {
            'Pit Cave Peak': [
                ('nosettings', 'inverted'),
                ('item', 'mirror')],
            'East Upper Mount Hebra': [('settings', 'inverted')]}
    },
    'Turtle Rock Portal (LW)': {
        'type': 'area',
        'link': {
            'Turtle Rock Portal (DW)': [('and', [
                ('settings', 'inverted'), ('item', 'mirror')])],
            'East Upper Mount Hebra': [
                ('nosettings', 'inverted'), ('item', 'hammer')],
            'Turtle Rock Portal': [('item', 'hammer')]}
    },
    'Turtle Rock Portal': {
        'type': 'area',
        'link': {
            'Turtle Rock Portal (DW)': [('and', [
                ('nosettings', 'inverted'), ('state', 'rabbit')])],
            'Turtle Rock Portal (LW)': [('and', [
                ('settings', 'inverted'), ('state', 'rabbit')])]}
    },
}



'''
Thieves' Town
'''

__all__ = 'LOCATIONS',


LOCATIONS = {
    "Thieves' Town Surface": {
        'type': 'area',
        'link': {
            'Kakariko': [('and', [
                ('nosettings', 'inverted'), ('item', 'mirror')])],
            'South Light World': [('and', [
                ('nosettings', 'inverted'), ('item', 'mirror')])],
            'Dark Bottle Ocarina': [],
            'South Dark World': [],
            'Frog Prison': [('item', 'titansmitts')],
            "East Thieves' Town": [
                ('item', 'titansmitts'),
                ('or', [
                    ('and', [
                        ('glitch', 'overworld'), ('item', 'pegasus')]),
                    ('glitch', 'major')])],
            "Thieves' Town Southern House Entrance (E)": [('item', 'bombs')],
            "Thieves' Town Shop Entrance (E)": [('item', 'hammer')],
            "Thieves' Town Eastern House Entrance (E)": [],
            "Thieves' Town Underground Entrance (E)": [('rabbitbarrier', None)],
            "Thieves' Town Western House Entrance (E)": [],
            'Skull Woods Surface': [],
            'Kakariko Portal (DW)': [('item', 'titansmitts')],
            'North Dark World': []}
    },
    'Frog Prison': {
        'type': 'area',
        'link': {
            'South Light World': [('and', [
                ('nosettings', 'inverted'), ('item', 'mirror')])],
            "Thieves' Town Surface": [('item', 'titansmitts')]}
    },
    "Thieves' Town Southern House Entrance (E)": {
        'type': 'entrance_unique', 'map': 'dark', 'coord': (67, 386),
        'link': {
            "Thieves' Town Surface": [],
            "Thieves' Town Southern House Entrance (I)": []}
    },
    "Thieves' Town Southern House Entrance (I)": {
        'type': 'interior',
        'link': {
            "Thieves' Town Southern House Entrance (E)": [],
            'Sealed House': []}
    },
    'Sealed House': {
        'type': 'chest', 'map': 'dark', 'coord': (67, 383),
        'link': {
            "Thieves' Town Southern House Entrance (E)": []}
    },
    "Thieves' Town Shop Entrance (E)": {
        'type': 'entrance_shop', 'map': 'dark', 'coord': (129, 353),
        'link': {
            "Thieves' Town Surface": [('item', 'hammer')],
            "Thieves' Town Shop Entrance (I)": []}
    },
    "Thieves' Town Shop Entrance (I)": {
        'type': 'interior',
        'link': {
            "Thieves' Town Shop Entrance (E)": []}
    },
    "Thieves' Town Eastern House Entrance (E)": {
        'type': 'entrance_unique', 'map': 'dark', 'coord': (131, 320),
        'link': {
            "Thieves' Town Surface": [],
            "Thieves' Town Eastern House Entrance (I)": []}
    },
    "Thieves' Town Eastern House Entrance (I)": {
        'type': 'interior',
        'link': {
            "Thieves' Town Eastern House Entrance (E)": [],
            'Deserted House': []}
    },
    'Deserted House': {
        'type': 'chest', 'map': 'dark', 'coord': (131, 314),
        'link': {
            "Thieves' Town Eastern House Entrance (I)": []}
    },
    "Thieves' Town Western House Entrance (E)": {
        'type': 'entrance_unique', 'map': 'dark', 'coord': (28, 310),
        'link': {
            "Thieves' Town Surface": [],
            "Thieves' Town Western House Entrance (I)": []}
    },
    "Thieves' Town Western House Entrance (I)": {
        'type': 'interior',
        'link': {
            "Thieves' Town Western House Entrance (E)": [],
            'Chest Game': []}
    },
    'Chest Game': {
        'type': 'chest', 'map': 'dark', 'coord': (28, 304),
        'link': {
            "Thieves' Town Western House Entrance (I)": []}
    },

    "East Thieves' Town": {
        'type': 'area',
        'link': {
            'Kakariko': [('and', [
                ('nosettings', 'inverted'), ('item', 'mirror')])],
            'Above The Bat': [('and', [
                ('nosettings', 'inverted'), ('item', 'mirror')])],
            "Thieves' Town Surface": [('item', 'titansmitts')],
            'Locked Chest': [('access', 'Blacksmiths')],
            'Stake Cave Entrance (E)': [('item', 'hammer')]}
    },
    'Locked Chest': {
        'type': 'area',
        'link': {
            "East Thieves' Town": []}
    },
    'Stake Cave Entrance (E)': {
        'type': 'entrance_unique', 'map': 'dark', 'coord': (204, 401),
        'link': {
            "East Thieves' Town": [],
            'Stake Cave Entrance (I)': []}
    },
    'Stake Cave Entrance (I)': {
        'type': 'interior',
        'link': {
            'Stake Cave Entrance (E)': [],
            'Stake Cave': []}
    },
    'Stake Cave': {
        'type': 'drop', 'map': 'dark', 'coord': (204, 401),
        'link': {
            'Stake Cave Entrance (I)': []}
    },

    "Thieves' Town Underground Entrance (E)": {
        'type': 'entrance_dungeon', 'map': 'dark', 'coord': (77, 322),
        'link': {
            "Thieves' Town Surface": [],
            "Thieves' Town Underground Entrance (I)": []}
    },
    "Thieves' Town Underground Entrance (I)": {
        'type': 'interior',
        'link': {
            "Thieves' Town Underground Entrance (E)": [],
            "Thieves' Town Underground": [
                ('settings', 'placement_advanced'),
                ('and', [
                    ('or', [
                        ('settings', 'swordless'), ('item', 'sword')]),
                    ('item', 'bottle')])]}
    },
    "Thieves' Town Underground": {
        'type': 'area', 'dungeon': "Thieves' Town",
        'link': {
            "Thieves' Town Underground Entrance (I)": [],
            "Thieves' Town Map": [('rabbitbarrier', None)],
            "Thieves' Town Rupees": [('rabbitbarrier', None)],
            "Thieves' Town Compass": [('rabbitbarrier', None)],
            "Thieves' Town Big Key": [('rabbitbarrier', None)],
            "Thieves' Town Boss Hall": [('bigkey', "Thieves' Town")]}
    },
    "Thieves' Town Map": {
        'type': 'dungeonchest', 'dungeon': "Thieves' Town",
        'link': {
            "Thieves' Town Underground": []}
    },
    "Thieves' Town Rupees": {
        'type': 'dungeonchest', 'dungeon': "Thieves' Town",
        'link': {
            "Thieves' Town Underground": []}
    },
    "Thieves' Town Compass": {
        'type': 'dungeonchest', 'dungeon': "Thieves' Town",
        'link': {
            "Thieves' Town Underground": []}
    },
    "Thieves' Town Big Key": {
        'type': 'dungeonchest', 'dungeon': "Thieves' Town",
        'link': {
            "Thieves' Town Underground": []}
    },
    "Thieves' Town Boss Hall": {
        'type': 'area', 'dungeon': "Thieves' Town",
        'link': {
            "Thieves' Town Underground": [],
            "Thieves' Town First Hidden Key": [('rabbitbarrier', None)],
            "Thieves' Town Boss": [('and', [
                ('access', "Thieves' Town Attic"),
                ('access', "Thieves' Town Blind's Cell")])],
            "Thieves' Town Gauntlet": [('smallkey', "Thieves' Town")]}
    },
    "Thieves' Town First Hidden Key": {
        'type': 'dungeonkey', 'dungeon': "Thieves' Town",
        'link': {
            "Thieves' Town Boss Hall": []}
    },
    "Thieves' Town Gauntlet": {
        'type': 'area', 'dungeon': "Thieves' Town",
        'link': {
            "Thieves' Town Boss Hall": [('rabbitbarrier', None)],
            "Thieves' Town Second Hidden Key": [('rabbitbarrier', None)],
            "Thieves' Town Attic": [('smallkey', "Thieves' Town")],
            "Thieves' Town Dungeons": []}
    },
    "Thieves' Town Second Hidden Key": {
        'type': 'dungeonkey', 'dungeon': "Thieves' Town",
        'link': {
            "Thieves' Town Gauntlet": []}
    },
    "Thieves' Town Attic": {
        'type': 'area', 'dungeon': "Thieves' Town",
        'link': {
            "Thieves' Town Gauntlet": [],
            "Thieves' Town Bombs": [('rabbitbarrier', None)]}
    },
    "Thieves' Town Bombs": {
        'type': 'dungeonchest_nokey', 'dungeon': "Thieves' Town",
        'link': {
            "Thieves' Town Attic": []}
    },
    "Thieves' Town Dungeons": {
        'type': 'area', 'dungeon': "Thieves' Town",
        'link': {
            "Thieves' Town Gauntlet": [],
            "Thieves' Town Treasure Room": [('smallkey', "Thieves' Town")],
            "Thieves' Town Blind's Cell": [('bigkey', "Thieves' Town")]}
    },
    "Thieves' Town Treasure Room": {
        'type': 'area', 'dungeon': "Thieves' Town",
        'link': {
            "Thieves' Town Dungeons": [],
            "Thieves' Town Treasure": [('item', 'hammer')]}
    },
    "Thieves' Town Treasure": {
        'type': 'dungeonchest_nokey', 'dungeon': "Thieves' Town",
        'link': {
            "Thieves' Town Treasure Room": [('item', 'hammer')]}
    },
    "Thieves' Town Blind's Cell": {
        'type': 'area', 'dungeon': "Thieves' Town",
        'link': {
            "Thieves' Town Dungeons": [],
            "Thieves' Town Blind's Key": []}
    },
    "Thieves' Town Blind's Key": {
        'type': 'dungeonchest_nokey', 'dungeon': "Thieves' Town",
        'link': {
            "Thieves' Town Blind's Cell": []}
    },
    "Thieves' Town Boss": {
        'type': 'dungeonboss', 'dungeon': "Thieves' Town",
        'link': {
            "Thieves' Town Boss Item": [('and', [
                ('or', [
                    ('settings', 'placement_advanced'),
                    ('settings', 'swordless'),
                    ('and', [
                        ('item', 'sword'),
                        ('or', [
                            ('item', 'cape'), ('item', 'byrna')])])]),
                ('or', [
                    ('item', 'sword'), ('item', 'hammer'),
                    ('item', 'byrna'), ('item', 'somaria')])])]}
    },
    "Thieves' Town Boss Item": {
        'type': 'dungeonchest_nokey', 'dungeon': "Thieves' Town",
        'link': {
            "Thieves' Town Reward": []}
    },
    "Thieves' Town Reward": {
        'type': 'dungeonreward', 'dungeon': "Thieves' Town",
        'link': {
            "Thieves' Town Underground Entrance (I)": []}
    },
}



'''
Item display
'''

import tkinter as tk
import tkinter.ttk as ttk
import typing

from ..config import CONFIG
from .. import items

from . import misc

__all__ = 'start',


class ItemWindow(tk.Toplevel):
    '''
    Inventory item display

    Instance variables:
        tracker: item tracker object
        layout: item layout in display
        helpertext: helper text variable
        scaling: scaling factor of individual buttons
        buttons: {'item identifier': button object}
    '''

    def __init__(self, tracker: items.ItemTracker):
        '''
        Args:
            tracker: item tracker object
        '''

        super().__init__()
        self.title('Items')

        self.tracker = tracker

        self.frame = ttk.Frame(self)
        self.frame.grid(column=0, row=0, sticky=misc.A)
        self.helpertext = tk.StringVar()
        self.helper = ttk.Label(
            self, textvariable=self.helpertext)
        self.helper.grid(column=0, row=1, sticky=tk.S)

        self.scaling = _scale_factors()

        self.buttons = {}
        for item in self.tracker:
            if (
                    self.tracker[item].location and
                    self.tracker[item].displayname and
                    self.tracker[item].icon):
                self._item_display(self.tracker[item])

    def _item_display(self, item) -> None:
        '''
        Make and place single item display object.

        Args:
            item: item object
        Writes:
            buttons
        '''

        button = ItemButton(item, self.frame, self.scaling)
        button.bind(
            '<Enter>', lambda _: self.helpertext.set(item.display()))
        button.bind(
            '<Leave>', lambda _: self.helpertext.set(''))
        button.bind('<ButtonRelease-1>', item.increase)
        button.bind('<ButtonRelease-3>', item.decrease)
        button.bind(
            '<ButtonRelease-1>',
            lambda _: self.helpertext.set(item.display()),
            add='+')
        button.bind(
            '<ButtonRelease-3>',
            lambda _: self.helpertext.set(item.display()),
            add='+')
        button.bind(
            '<ButtonRelease-1>',
            lambda _: self.tracker.save(),
            add='+')
        button.bind(
            '<ButtonRelease-3>',
            lambda _: self.tracker.save(),
            add='+')
        button.bind(
            '<ButtonRelease-1>',
            lambda _: button.check_state(item),
            add='+')
        button.bind(
            '<ButtonRelease-3>',
            lambda _: button.check_state(item),
            add='+')
        button.grid(column=item.location[1], row=item.location[0])
        self.buttons[item.identifier] = button
        

    def reset(self) -> None:
        '''
        Reset items to default.
        '''

        self.tracker.reset()
        for item in self.tracker:
            if self.tracker[item].icon:
                self.buttons[item].check_state(self.tracker[item])


class ItemButton(tk.Canvas):
    '''
    Single item display object.
    '''

    def __init__(
            self, item, parent: ttk.Frame, scaling: (int, int)):
        '''
        item: item object
        parent: parent widget for object
        scaling: button up- and downscale
        '''

        super().__init__(
            parent, height=64 * scaling[0] / scaling[1],
            width=64 * scaling[0] / scaling[1])
        self.scaling = scaling
        self.img = None
        self.icon = None
        self.check_state(item)

    def check_state(self, item) -> None:
        '''
        Check button state and make adjustments if necessary.

        Args:
            item: item object
        '''        

        self.delete(self.img)
        icon = tk.PhotoImage(
            file=item.icon[item.index()], master=self.master)
        if self.scaling[0] != 1:
            icon = icon.zoom(self.scaling[0], self.scaling[0])
        if self.scaling[1] != 1:
            icon = icon.subsample(self.scaling[1], self.scaling[1])
        if not item.state():
            for x in range(icon.width()):
                for y in range(icon.height()):
                    bw = sum(icon.get(x, y)) // item.disabled[0]
                    if bw in (0, 255):
                        continue
                    bw *= item.disabled[1]
                    bw = 255 if bw > 255 else int(bw)
                    icon.put('#{0:02x}{0:02x}{0:02x}'.format(bw), (x, y))
        self.img = self.create_image(0, 0, anchor=tk.NW, image=icon)
        self.icon = icon


def _scale_factors() -> (int, int):
    '''
    Calculate up- and downscale factor.

    Returns:
        int: upscale factor
        int: downscale factor
    '''

    scaling = CONFIG['icon_size']
    for up in range(1, 1000):
        if not (scaling * up) % 1:
            upscale = int(scaling * up)
            downscale = up
            break
    else:
        CONFIG.set('icon_size', 1)
        assert False
    return upscale, downscale


def start(worldtracker) -> ItemWindow:
    '''
    Initialise item tracker window.

    Args:
        worldtracker: world state tracker
    '''

    return ItemWindow(items.ItemTracker(worldtracker))



'''
Item tracking object.
'''

import operator
import os.path
import typing

from ..config.images import image

__all__ = 'ItemObj',


class ItemObj(object):
    '''
    Inventory item

    Progressive items are listed as one.

    Instance variables:
        identifier: item identifier string
        location: (column, row) location on display
        length: number of item progressions
        displayname: name(s) displayed in UI
        icon: path to image file(s) associated with item
        disabled: black&white conversion
        link: items (and requirement) linked with this item
        linkitems: item objects linked to this item
        default: default numer of items in inventory
        inventory: current number of items in inventory
        tracker: world state tracker
    '''

    def __init__(self, identifier: str, location: typing.Sequence[int],
                 length: int, displayname: typing.Sequence[str],
                 icon: typing.Sequence[str], disabled: typing.Sequence[int],
                 links: typing.Mapping[str, typing.Sequence[int]],
                 default: int, tracker):
        '''
        Args:
            identifier: internal item name
            location: (row, column) of item on tracker GUI
            length: number of progressions in item
            displayname: name(s) of item displayed on tracker GUI
            icon: icon(s) used on tracker GUI
            disabled: black&white conversion
            links: items linked with posession of item
            default: initial item progression
            tracker: world state tracker
        '''

        self.identifier = identifier
        self.location = location
        self.length = length
        self.displayname = displayname
        self.icon = tuple(image(i) for i in icon)
        self.disabled = disabled
        self.link = links
        self.linkitems = {}
        self.default = default
        self.inventory = default
        self.tracker = tracker
        self.restore_inventory(self.default)

    def index(self) -> int:
        '''
        Return current displayname/image index.

        Returns:
            int: index used for sequence attributes
        '''

        idx = self.inventory if self.inventory < 1 else self.inventory - 1
        return idx

    def display(self) -> str:
        '''
        Return currently applicable item display string.

        Returns:
            str: name to be displayed in application
        '''

        idx = self.index()
        item_name = self.displayname[idx]
        return item_name

    def state(self) -> bool:
        '''
        Return current state of item.

        Returns:
            str: True if item is active, else False
        '''

        return self.inventory > 0

    def increase(self, *args) -> None:
        '''
        Left-click on item
        '''

        if self.inventory < self.length:
            self.inventory += 1
        self.update_links()
        self.tracker.set_item(self.identifier, self.inventory)

    def decrease(self, *args) -> None:
        '''
        Right-click on item
        '''
        
        if self.inventory > 0:
            self.inventory -= 1
        self.update_links()
        self.tracker.set_item(self.identifier, self.inventory)

    def reset(self) -> None:
        '''
        Reset item.
        '''

        to_remove = self.inventory - self.default
        if to_remove > 0:
            for _ in range(to_remove):
                self.decrease(_)
        elif to_remove < 0:
            for _ in range(-to_remove):
                self.increase(_)

    def restore_inventory(self, quantity: int) -> None:
        '''
        Set inventory number.

        Args:
            quantity: number to set inventory to
        '''

        self.inventory = quantity
        self.update_links()
        self.tracker.set_item(self.identifier, quantity)

    def add_links(linkitems: typing.Sequence) -> None:
        '''
        Register linked items.

        Args:
            linkitems: list of item objects
        '''

        for li in linkitems:
            self.linkitems[li.identifier] = li

    def update_links(self) -> None:
        '''
        Update state of linked items according to current inventory.
        '''

        for li in self.linkitems:
            if self.inventory in self.link[li]:
                self.linkitems[li].restore_inventory(1)
            else:
                self.linkitems[li].restore_inventory(0)



'''
World state tracker
'''

import typing

from ..config import CONFIG
from ..dungeons import lists as dungeonlists
from .. import entrances
from .. import maps
from .. import ruleset

__all__ = 'Tracker',


class _DelayCheck(Exception):
    '''
    Thrown by Tracker.parse_requirement() when a check should be delayed.
    '''

    def __init__(self, delayclass: str):
        '''
        Args:
            delayclass: 'common', 'smallkey', 'bigkey', 'reward', 'boss'
        '''

        super().__init__()
        assert delayclass in (
            'common', 'smallkey', 'bigkey', 'reward', 'maybe', 'boss')
        self.delayclass = delayclass

class _SmallKeyCheck(Exception):
    '''
    Raised when small key door is encountered.
    '''

    def __init__(self, dungeon: str):
        '''
        Args:
            requirement: number of required key chests
        '''

        super().__init__()
        self.dungeon = dungeon


class _BigKeyCheck(Exception):
    '''
    Raised when big key lock is encountered.
    '''

    def __init__(self, dungeon):
        '''
        Args:
            requirement: number of required key chests
        '''

        super().__init__()
        self.dungeon = dungeon


class _RewardCheck(Exception):
    '''
    Raised when reward is required.
    '''


class Tracker(object):
    '''
    World state tracker

    Instance variables:
        items: item inventory
        dungeons: dungeon state
        maps: {'light', 'dark'} world map displays
        locationtracker: location tracker
        ruleset: location rules
        settings: game settings
        startpoint: starting locations
        entrances: entrance tracker
        crystals: crystal requirements for {'tower', 'ganon'}
        keys = {'dungeon': {'small': int, 'big': bool}}
    '''

    def __init__(self):
        self.items = {}
        self.dungeons = {}
        self.maps = {'light': None, 'dark': None}
        self.locationtracker = maps.LocationTracker()
        self.entrances = entrances.EntranceTracker(self)
        self.crystals = {'tower': -1, 'ganon': -1}
        self.keys = {}
        self.update_rules()

    def update_rules(self) -> None:
        '''
        Retrieve location rules.
        '''

        self.ruleset = ruleset.Ruleset()
        self._make_settings()
        for dungeon in self.ruleset.dungeons:
            if (dungeon not in self.keys or 'small' not in self.keys[dungeon] or
                'big' not in self.keys[dungeon]):
                self.keys[dungeon] = {'small': 0, 'big': False}
        self.rulecheck()

    def _make_settings(self) -> None:
        '''
        Build game settings.
        '''

        self.settings = set()
        if CONFIG['entrance_randomiser']:
            self.settings.add('entrance')
        self.settings.add(CONFIG['world_state'].lower())
        if 'standard' in self.settings:
            self.settings.add('open')
        self.settings.add('glitches_{0:s}'.format(
            {'None': 'none', 'Overworld Glitches': 'overworld',
             'Major Glitches': 'major'}[CONFIG['glitch_mode']]))
        self.settings.add('placement_{0:s}'.format(
            CONFIG['item_placement'].lower()))
        if CONFIG['dungeon_items'] in (
                'Maps/Compasses', 'Maps/Compasses/Small Keys', 'Keysanity'):
            self.settings.update(('random_map', 'random_compass'))
        if CONFIG['dungeon_items'] in (
                'Maps/Compasses/Small Keys', 'Keysanity'):
            self.settings.add('random_smallkey')
        if CONFIG['dungeon_items'] == 'Keysanity':
            self.settings.add('random_bigkey')
        self.settings.add('goal_{0:s}'.format(
            {'Defeat Ganon': 'ganon', 'Fast Ganon': 'fastganon',
             'All Dungeons': 'dungeons', 'Master Sword Pedestal': 'pedestal',
             'Triforce Hunt': 'triforce'}[CONFIG['goal']]))
        if CONFIG['swords'] == 'Swordless':
            self.settings.add('swordless')
        if CONFIG['enemiser']:
            self.settings.add('enemiser')
        if CONFIG['majoronly']:
            self.settings.add('majoronly')
        self._set_startpoint()

    def _set_startpoint(self) -> None:
        '''
        Choose starting point.
        '''

        self.startpoint = ["Link's House"]
        if 'inverted' in self.settings:
            self.startpoint.append('Dark Chapel Entrance (I)')
        else:
            self.startpoint.append('Sanctuary')

    def reapply(self) -> None:
        '''
        Purge map and recreate with (possibly) updated settings.
        '''

        self.update_rules()
        for maptype in self.maps:
            self.maps[maptype].remove_buttons()
            self.make_buttons(maptype)
        if 'entrance' in self.settings:
            self.ruleset.disconnect_entrances()
            self.maps['light'].set_entrance()
        else:
            self.ruleset = ruleset.Ruleset()
        self.rulecheck()

    def set_item(self, itemname: str, itemnumber: int) -> None:
        '''
        Set item inventory.

        Args:
            itemname: item identifier
            inventory: new item inventory to overwrite old one
        '''

        self.items[itemname] = itemnumber
        self.rulecheck()

    def set_dungeon(self, dungeonname: str, dungeondata: dict) -> None:
        '''
        Set dungeon data.

        Args:
            dungeonname: dungeon name
            dungeondata: new dungeon data
        '''

        self.dungeons[dungeonname] = dungeondata
        self.rulecheck()

    def add_map(self, maptype: str, mapdisplay) -> None:
        '''
        Add map to tracker.

        Args:
            maptype: 'light' or 'dark'
            mapdisplay: map display object
        '''

        assert maptype in ('light', 'dark')
        self.maps[maptype] = mapdisplay
        self.maps[maptype].add_location_tracker(self.locationtracker)
        if all(m is not None for m in self.maps.values()):
            self.maps['light'].other = self.maps['dark']
            self.maps['dark'].other = self.maps['light']
        self.make_buttons(maptype)
        if 'entrance' in self.settings:
            self.ruleset.disconnect_entrances()
            if self.maps['light'] and self.maps['dark']:
                self.maps['light'].set_entrance()
        self.maps[maptype].add_entrance_tracker(self.entrances)
        self.rulecheck()

    def make_buttons(self, maptype: str) -> None:
        '''
        Send button data to map display.

        Args:
            maptype: 'light' or 'dark'
        '''

        assert maptype in ('light', 'dark')
        if CONFIG['entrance_randomiser']:
            gametype = 'entrance'
        elif CONFIG['world_state'] == 'Retro':
            gametype = 'item_retro'
        else:
            gametype = 'item'
        self.maps[maptype].place_buttons(
            self.ruleset.locations(
                gametype, maptype, 'majoronly' in self.settings))

    def update_buttons(self, available: typing.Mapping,
                       visible: typing.Sequence[str]) -> None:
        '''
        Send updated availability info to map displays.

        Args:
            available: {'available location': 'type of availability'}
            visible: list of visible locations
        '''

        # Go through maps.
        for mapd in self.maps.values():

            # Go through buttons:
            for button in mapd.button:

                # Check dungeon status.
                if mapd.button[button]['type'] == 'dungeon':
                    dungeonchecks = []
                    for loc in self.ruleset.dungeons[button]:
                        if (self.ruleset[loc].type.startswith('dungeonchest')
                            and not loc.endswith(' Boss Item')):
                            dungeonchecks.append(loc in available)
                    if all(dungeonchecks):
                        try:
                            clearable = available[
                                '{0:s} Big Key'.format(button)]
                        except KeyError:  # Castle Tower
                            clearable = available[
                                '{0:s} Second Chest'.format(button)]
                    elif any(dungeonchecks):
                        clearable = 'visible'
                    else:
                        clearable = 'unavailable'
                    if '{0:s} Reward'.format(button) in available:
                        finishable = available['{0:s} Boss'.format(button)]
                        if (finishable == 'available' and
                            clearable == 'indirect'):
                            finishable = 'indirect'
                    else:
                        finishable = 'unavailable'
                    state = [clearable, finishable]
                    mapd.set_dungeon_availability(button, state)

                # Check location status.
                else:
                    if button in available:
                        state = available[button]
                    elif button in visible:
                        state = 'visible'
                    else:
                        state = 'unavailable'
                    mapd.set_availability(button, state)

    def rulecheck(self) -> None:
        '''
        Update availability of all locations.
        '''

        # Don't run if both maps aren't yet available.
        if not all(self.maps.values()):
            return

        # Create fresh rules.
        for loc in self.ruleset.values():
            loc.checked = False
        keys = {}
        for dungeon in self.ruleset.dungeons:
            if self.ruleset[dungeon].type == 'dungeon':
                keys[dungeon] = {'small': self.keys[dungeon]['small'],
                                 'big': self.keys[dungeon]['big']}
            else:
                keys[dungeon] = {'small': 1, 'big': False}

        # Prepare rulecheck.
        available = {}
        reachable = [_Connection(s) for s in self.startpoint]
        visible = set()
        delayed = {'common': [], 'smallkey': [], 'bigkey': [], 'boss': [],
                   'reward': [], 'maybe': []}
        keychecked = {loc: set() for loc in self.ruleset.dungeons}
        hadkey = dict.fromkeys(self.ruleset.dungeons.keys(), False)
        maybedungeons = set()

        # Go through locations.
        while reachable:

            # Retrieve next available location.
            current, state = reachable.pop(0).get()

            # Mark current location.
            if self.ruleset[current].checked is not None:
                if 'maybe' in state:
                    available[current] = 'maybe'
                elif 'boss' in state:
                    available[current] = 'indirect'
                else:
                    available[current] = 'available'
            self.ruleset[current].checked = True

            # Add fixed keys.
            if self.ruleset[current].type == 'dungeonkey':
                keys[self.ruleset[current].dungeon]['small'] += 1
                hadkey[self.ruleset[current].dungeon] = True

            # Add chest keys.
            if (not 'random_smallkey' in self.settings and
                self.ruleset[current].type.startswith('dungeonchest') and
                self.ruleset[self.ruleset[current].dungeon].type == 'dungeon'):
                keypools = self.ruleset[self.ruleset[current].dungeon].keypools
                spheres_defined = False
                for sphere in keypools:
                    if (sphere['type'] != 'small' or
                        current not in sphere['chests']):
                        continue
                    for s in sphere['settings']:
                        if s not in self.settings:
                            break
                    else:
                        spheres_defined = True
                        remaining = sum(
                            loc not in keychecked[self.ruleset[current].dungeon]
                            for loc in sphere['chests'])
                        if remaining <= sphere['keys']:
                            keys[self.ruleset[current].dungeon]['small'] += 1
                            hadkey[self.ruleset[current].dungeon] = True
                            break
                if not spheres_defined:
                    total = self.ruleset.dungeons[
                        self.ruleset[current].dungeon].keylocations()
                    if current in total:
                        remaining = sum(
                            loc not in keychecked[self.ruleset[current].dungeon]
                            for loc in total)
                        totalkeys = dungeonlists.DUNGEONS[
                            self.ruleset[current].dungeon][1]
                        if remaining <= totalkeys:
                            keys[self.ruleset[current].dungeon]['small'] += 1
                            hadkey[self.ruleset[current].dungeon] = True
                keychecked[self.ruleset[current].dungeon].add(current)

            # Go through links.
            for link in self.ruleset[current].link:

                # Don't recheck already visited locations.
                if self.ruleset[link].checked:
                    continue

                # Parse requirement.
                newstate = state.copy()
                try:
                    ret, addstate = self._parse_requirement(
                        self.ruleset[current].link[link], state, None, keys)
                except _DelayCheck as err:
                    newlink = (link, _Connection(
                        self.ruleset[current].link[link], state))
                    if newlink not in delayed[err.delayclass]:
                        delayed[err.delayclass].append(newlink)
                    ret = False
                if ret:
                    for retstate in addstate:
                        s = retstate.split(';')
                        if len(s) > 1 and s[1] == 'add':
                            newstate.add(s[0])
                        elif s[0] in newstate:
                            newstate.remove(s[0])
                        elif not (len(s) > 1 and s[1] == 'dis'):
                            newstate.add(s[0])
                    newlink = _Connection(link, newstate)
                    if newlink not in reachable:
                        reachable.append(newlink)

            # Check visible locations.
            for link in self.ruleset[current].visible:
                if self._parse_requirement(
                        self.ruleset[current].visible[link], state, None,
                        keys)[0]:
                    visible.add(link)

            # Go though delayed checks.
            ret = False
            addstate = []
            if not reachable:
                for delayclass in delayed:
                    for idx, loc in enumerate(delayed[delayclass]):
                        current, state = loc[1].get()
                        newstate = state.copy()
                        try:
                            ret, addstate = self._parse_requirement(
                                current, state, available, keys)
                        except _SmallKeyCheck as err:
                            ret = self._check_smallkeys(
                                available, err.dungeon, keys)
                            if not ret and hadkey[err.dungeon]:
                                ret = True
                                addstate = ('maybe;add',)
                                maybedungeons.add(err.dungeon)
                        except _BigKeyCheck as err:
                            avail, maybe = self._check_bigkey(
                                available, err.dungeon)
                            ret = False
                            if avail or maybe:
                                ret = True
                            if ((not avail and maybe) or
                                err.dungeon in maybedungeons):
                                addstate = ('maybe;add',)
                        if ret:
                            for retstate in addstate:
                                s = retstate.split(';')
                                if len(s) > 1 and s[1] == 'add':
                                    newstate.add(s[0])
                                elif s[0] in state:
                                    newstate.remove(s[0])
                                elif not (len(s) > 1 and s[1] == 'dis'):
                                    newstate.add(s[0])
                            newlink = _Connection(
                                delayed[delayclass].pop(idx)[0], newstate)
                            if newlink not in reachable:
                                reachable.append(newlink)
                            break
                    if reachable:
                        break

        # Send updated availability to map displays.
        self.update_buttons(available, visible)

    def _parse_requirement(
            self, req: typing.Sequence[typing.Sequence], state: typing.Sequence,
            nodelay: typing.Sequence[str], keys: dict = {},
            collector=any) -> (bool, list):
        '''
        Parse generic link requirement object.

        When more than one requirement is given, this function defaults to OR!

        Args:
            req: [('requirement type', requirement object)]
            state: current connection state
            nodelay: don't throw DelayCheck when encountering access
               requirement, instead check against this list of locations
            keys: current number of available keys
            collector: something like any() or all()
        Returns:
            bool: True if requirements are met
            list: list of state flags to toggle
        '''

        if not req:
            return True, []
        result = []
        addstate = []

        for rtype, robj in req:

            # OR
            if rtype == 'or':
                sub = self._parse_requirement(robj, state, nodelay, keys)
                result.append(sub[0])
                if sub[0]:
                    addstate.extend(sub[1])

            # AND
            elif rtype == 'and':
                sub = self._parse_requirement(robj, state, nodelay, keys, all)
                result.append(sub[0])
                if sub[0]:
                    addstate.extend(sub[1])

            # settings
            elif rtype == 'settings':
                result.append(robj in self.settings)

            # nosettings
            elif rtype == 'nosettings':
                result.append(robj not in self.settings)

            # item
            elif rtype == 'item':
                rabbititems = (
                    'mushroom', 'lantern', 'mudora', 'bottle', 'mirror')
                if ('rabbit' in state and (
                        not 'pearl' in self.items or self.items['pearl'] == 0)
                    and robj not in rabbititems):
                    result.append(False)
                else:
                    result.append(robj in self.items and self.items[robj] > 0)
                    if robj == 'mirror':
                        addstate.append('rabbit;dis')

            # access
            elif rtype == 'access':
                if not nodelay:
                    raise _DelayCheck('common')
                result.append(robj in nodelay)
                if result[-1] and nodelay[robj] in ('maybe', 'indirect'):
                    addstate.append('{0:s};add'.format(
                        'maybe' if nodelay[robj] == 'maybe' else 'boss'))

            # glitch
            elif rtype == 'glitch':
                if robj == 'overworld':
                    result.append('glitches_overworld' in self.settings or
                                  'glitches_major' in self.settings)
                elif robj == 'major':
                    result.append('glitches_major' in self.settings)

            # medallion
            elif rtype == 'medallion':
                if 'rabbit' in state and not 'pearl' in self.items:
                    result.append(False)
                res, maybe = self._check_medallion(robj)
                result.append(res)
                if maybe:
                    addstate.append('maybe;add')

            # mudora
            elif rtype == 'mudora':
                assert robj in ('take', 'see')
                result.append(
                    'mudora' in self.items and self.items['mudora'] > 0 and
                    (('swordless' in self.settings and
                      'hammer' in self.items and self.items['hammer'] > 0) or
                     'mastersword' in self.items and
                     self.items['mastersword'] > 0))

            # pendant/crystals
            elif rtype in ('pendant', 'crystals'):
                if not nodelay:
                    raise _DelayCheck('reward')
                res, flag = self._reward_locations(robj, nodelay)
                result.append(res)
                if flag:
                    addstate.append('{0:s};add'.format(flag))

            # smallkey
            elif rtype == 'smallkey':
                if 'retro' in self.settings:
                    result.append(True)
                elif not nodelay:
                    raise _DelayCheck('smallkey')
                else:
                    if 'random_smallkey' in self.settings:
                        if keys[robj]['small'] > 0:
                            keys[robj]['small'] -= 1
                            result.append(True)
                        else:
                            result.append(False)
                    else:
                        raise _SmallKeyCheck(robj)

            # bigkey
            elif rtype == 'bigkey':
                if not nodelay:
                    raise _DelayCheck('bigkey')
                else:
                    if 'random_bigkey' in self.settings:
                        if keys[robj]['big']:
                            result.append(True)
                        else:
                            result.append(False)
                    else:
                        raise _BigKeyCheck(robj)

            # macro
            elif rtype == 'macro':
                if robj == 'ganon':
                    if 'goal_ganon' in self.settings:
                        sub = self._parse_requirement(
                                [('and', [
                                    ('crystals', 'ganon'),
                                    ('macro', 'ganondrop')])],
                                state, nodelay, keys, all)
                    elif 'goal_fastganon' in self.settings:
                        sub = self._parse_requirement(
                                [('crystals', 'ganon')], state, nodelay, keys)
                    elif 'goal_dungeons' in self.settings:
                        sub = self._parse_requirement(
                                [('and', [
                                    ('crystals', 'ganon'),
                                    ('pendant', 'courage'),
                                    ('pendant', 'power'),
                                    ('pendant', 'wisdom')])],
                                state, nodelay, keys, all)
                    elif 'goal_pedestal' in self.settings:
                        sub = self._parse_requirement(
                                [('and', [
                                    ('pendant', 'courage'),
                                    ('pendant', 'power'),
                                    ('pendant', 'wisdom')])],
                                state, nodelay, keys, all)
                    elif 'goal_triforce' in self.settings:
                        sub = self._parse_requirement(
                                [('and', [
                                    ('crystals', 'ganon'),
                                    ('pendant', 'courage'),
                                    ('pendant', 'power'),
                                    ('pendant', 'wisdom')])],
                                state, nodelay, keys, all)
                    else:
                        assert False
                elif robj == 'ganonstower':
                    sub = self._parse_requirement(
                        [('crystals', 'ganonstower')], state, nodelay, keys)
                elif robj == 'ganondrop':
                    if not nodelay:
                        raise _DelayCheck('reward')
                    if self._check_dungeon_state("Ganon's Tower"):
                        sub = (True, ())
                    else:
                        sub = self._parse_requirement(
                            [('access', "Ganon's Tower Reward")],
                            state, nodelay, keys)
                        sub[1].append('boss;add')
                else:
                    print(rtype, robj)
                    assert False
                result.append(sub[0])
                if sub[0]:
                    addstate.extend(sub[1])

            # state
            elif rtype == 'state':
                statestr = robj.split(';')
                assert statestr[0] in ('rabbit', 'maybe')
                if statestr[0] == 'maybe' and not nodelay:
                    raise _DelayCheck('maybe')
                addstate.append(statestr[0])
                result.append(True)

            # rabbitbarrier
            elif rtype == 'rabbitbarrier':
                if 'rabbit' in state and not self._parse_requirement(
                        [('item', 'pearl')], state, nodelay)[0]:
                    result.append(False)
                else:
                    result.append(True)

            # boss
            elif rtype == 'boss':
                if not nodelay:
                    raise _DelayCheck('boss')
                if not self._check_dungeon_state(robj):
                    addstate.append('boss;add')
                result.append(True)

            # error
            else:
                print(rtype, robj)
                assert False

        return collector(result), addstate

    def _check_smallkeys(
            self, availability: typing.Sequence[str], dungeon: str,
            keys: dict) -> bool:
        '''
        Check number of available small keys.

        Args:
            availability: list of already available locations
            dungeon: dungeon name
            keys: {dungeon name: number of available keys}
        Returns:
            bool: True if key requirement is met
        '''

        if 'random_smallkey' in self.settings:
            addkeys = self.keys[dungeon]['small']
        else:
            addkeys = 0
        if keys[dungeon]['small'] + addkeys > 0:
            keys[dungeon]['small'] -= 1
            return True
        return False

    def _check_bigkey(
            self, availability: typing.Sequence[str],
            dungeon: str) -> (bool, bool):
        '''
        Check number of available small keys.

        Args:
            availability: list of already available locations
            dungeon: dungeon name
        Returns:
            bool: True if key requirement is met
            bool: True if key requirement might be met
        '''

        if 'random_bigkey' in self.settings:
            return self.keys[dungeon]['big']
        for pool in self.ruleset[dungeon].keypools:
            if pool['type'] == 'big':
                bigpool = pool['chests']
                break
        else:
            bigpool = {
                loc for loc in self.ruleset.dungeons[dungeon]
                if self.ruleset[loc].type == 'dungeonchest'}
        avail = tuple(loc in availability for loc in bigpool)
        ret = all(avail)
        maybe = any(avail)
        return ret, maybe

    def _check_medallion(self, dungeon) -> (bool, bool):
        '''
        Check whether medallion requirement is sattisfied.

        Args:
            dungeon: 'Misery Mire' or 'Turtle Rock'
        Returns:
            bool: True if fight is finishable
            bool: True if 'maybe' state should be added
        '''

        assert dungeon in ('Misery Mire', 'Turtle Rock')
        req = self.dungeons[dungeon]['medallion']
        if req == 'unknown':
            res = (self.items['bombos'] > 0, self.items['ether'] > 0,
                   self.items['quake'] > 0)
            if all(res):
                res = True, False
            elif any(res):
                res = True, True
            else:
                res = False, False
        else:
            res = self.items[req] > 0, False
        if not 'swordless' in self.settings and self.items['sword'] == 0:
            res = False, False
        return res

    def _reward_locations(
            self, reward: str, available: typing.Sequence[str]) -> (bool, str):
        '''
        Retrieve required locations for reward.

        Args:
            reward: 'courage', 'power', 'wisdom', 'fairy', 'ganonstower',
                'ganon'
            available: list of available locations
        Returns:
            bool: True if all required locations are available
            str: '', 'maybe' or 'boss'
        '''

        # Sanity check
        assert reward in (
            'courage', 'power', 'wisdom', 'fairy', 'ganonstower', 'ganon')

        # Conversion between ruleset files and reward names.
        conversion = {
            'courage': 'courage', 'power': 'powerwisdom',
            'wisdom': 'powerwisdom', 'fairy': '56crystal',
            'ganonstower': 'crystal', 'ganon': 'crystal'}

        # Get current dungeons for desired reward type.
        rewarddungeons = []
        knowndungeons = []
        for dungeon in self.dungeons:
            if not 'reward' in self.dungeons[dungeon]['features']:
                continue
            if self.dungeons[dungeon]['reward'] == conversion[reward]:
                rewarddungeons.append(dungeon)
                knowndungeons.append(dungeon)
            if (reward.startswith('ganon') and
                self.dungeons[dungeon]['reward'] == '56crystal'):
                rewarddungeons.append(dungeon)
                knowndungeons.append(dungeon)

        # Check whether all required reward locations are known.
        required = {
            'courage': 1, 'power': 2, 'wisdom': 2, 'fairy': 2,
            'ganonstower': self.crystals['tower'],
            'ganon': self.crystals['ganon']}
        if required[reward] < 0:
            unknown_requirement = True
        else:
            unknown_requirement = False
        total = dict.fromkeys(
            ('courage', 'powerwisdom', '56crystal', 'crystal'), 0)
        for dungeon in self.dungeons:
            if self.dungeons[dungeon]['reward'] != 'unknown':
                total[self.dungeons[dungeon]['reward']] += 1
                if self.dungeons[dungeon]['reward'] == '56crystal':
                    total['crystal'] += 1

        # Check if enough rewards are known.
        enough = len(rewarddungeons) >= required[reward]

        # If not enough, count including dungeons with unknown reward.
        if not enough:
            rewarddungeons = []
            for dungeon in self.dungeons:
                if not 'reward' in self.dungeons[dungeon]['features']:
                    continue
                if self.dungeons[dungeon]['reward'] in (
                        conversion[reward], 'unknown'):
                    rewarddungeons.append(dungeon)
                if (reward.startswith('ganon') and
                    self.dungeons[dungeon]['reward'] == '56crystal'):
                    rewarddungeons.append(dungeon)

        # Check if there is enough now.
        locations = []
        maybes = 0
        boss_required = 0
        for dungeon in rewarddungeons:
            if self._check_dungeon_state(dungeon):
                locations.append(True)
                continue
            res, add = self._parse_requirement(
                [('access', '{0:s} Reward'.format(dungeon))], [], available)
            locations.append(res)
            for retstate in add:
                if res and retstate.split(';')[0] == 'maybe':
                    maybes += 1
                    break
            boss_required += 1

        # Add all information together.
        addflag = ''
        if enough:
            ret = sum(locations) >= required[reward]
            if required[reward] + boss_required > total[conversion[reward]]:
                addflag = 'boss'
            if required[reward] + maybes > total[conversion[reward]]:
                addflag = 'maybe'
            if unknown_requirement:
                if sum(locations) - maybes < 7:
                    addflag = 'maybe'
                elif boss_required:
                    addflag = 'boss'
        else:
            ret = sum(locations) >= required[reward]
            if not ret:
                ret = unknown_requirement
                addflag = 'maybe'
            else:
                if all(locations):
                    addflag = 'boss'
                else:
                    addflag = 'maybe'
        return ret, addflag

    def total_chests(self, dungeon: str) -> int:
        '''
        Retrieve number of chests available in dungeon.

        Args:
            dungeon: name of dungeon
        Returns:
            int: number of non-dungeon-specific items
        '''

        total_chests = 0
        for locname in self.ruleset.dungeons[dungeon]:
            majorcheck = (
                'majoronly' not in self.settings or
                locname in ruleset.MAJORLOCATIONS)
            loc = self.ruleset.dungeons[dungeon][locname]
            if loc.type.startswith('dungeonchest') and majorcheck:
                total_chests += 1
        return total_chests

    def set_dungeon_state(self, dungeon: str, state: [bool, bool]) -> None:
        '''
        Set whether dungeon has been checked or not.

        Args:
            dungeon: name of dungeon
            state: (all items collected, boss defeated)
        '''

        assert isinstance(state, list)
        assert len(state) == 2
        if dungeon in self.maps['light'].button:
            mapd = 'light'
        else:
            mapd = 'dark'
        self.maps[mapd].tracker[dungeon] = state
        self.rulecheck()

    def _check_dungeon_state(self, dungeon: str) -> bool:
        '''
        Check whether dungeon reward has been collected or not.

        Args:
            dungeon: name of dungeon
        Returns:
            bool: True if dungeon reward has been picked up
        '''

        try:
            ret = self.maps['dark'].tracker[dungeon][1]
        except KeyError:
            ret = self.maps['light'].tracker[dungeon][1]
        return not ret

    def set_crystal_requirement(self, tower: int, ganon: int) -> None:
        '''
        Set crystal requirements for Tower and Ganon Fight access.

        Args:
            tower: crystals required for Ganon's Tower access
            ganon: crystals required for the Ganon Fight
        '''

        self.crystals['tower'] = tower
        self.crystals['ganon'] = ganon
        self.rulecheck()


class _Connection(object):
    '''
    Connected location

    Instance variables:
        name: name of connected location
        state: list of states linked to connection
    '''

    def __init__(self, init: str = '', state: typing.Sequence = set()):
        '''
        Args:
            init: name of connected location
            state: list of states linked to connection
        '''

        super().__init__()
        self.name = init
        self.state = state

    def __eq__(self, other):
        '''
        Equality check

        This will raise an error if 'other' is not '_Connection' type.
        '''

        return self.get() == other.get()

    def get(self) -> (str, tuple):
        '''
        Unpack data.

        Returns:
            str: name of connected location
            state: list of states linked to connection
        '''

        return self.name, self.state



'''
Program configuration infrastructure
'''

from . import config
from . import globals

__all__ = 'CONFIG', 'CONFIGDIRECTORY'

globals.CONFIG = config.Config()
CONFIG = globals.CONFIG

CONFIGDIRECTORY = config.config_directory()



'''
Dungeons
'''

__all__ = 'LOCATIONS',


LOCATIONS = {
    'Castle Dungeons': {
        'type': 'drop', 'map': 'light', 'coord': (332, 290),
        'keypools': [],
        'link': {
            "Zelda's Treasure": []}
    },
    'East Palace': {
        'type': 'dungeon', 'map': 'light', 'coord': (636, 251),
        'keypools': [],
        'link': {}
    },
    'Desert Palace': {
        'type': 'dungeon', 'map': 'light', 'coord': (49, 528),
        'keypools': [
            {'type': 'small', 'settings': (), 'keys': 1,
             'chests': ('Desert Palace Map', 'Desert Palace Small Key')},
            {'type': 'big', 'settings': (),
             'chests': (
                 'Desert Palace Compass', 'Desert Palace Big Key',
                 'Desert Palace Map', 'Desert Palace Small Key')}],
        'link': {}
    },
    'Mountain Tower': {
        'type': 'dungeon', 'map': 'light', 'coord': (372, 14),
        'keypools': [
            {'type': 'small', 'settings': (), 'keys': 1,
             'chests': ('Mountain Tower Small Key', 'Mountain Tower Map')},
            {'type': 'big', 'settings': (),
             'chests': (
                 'Mountain Tower Small Key', 'Mountain Tower Map',
                 'Mountain Tower Big Key')}],
        'link': {}
    },
    'Castle Tower': {
        'type': 'dungeon', 'map': 'light', 'coord': (332, 244),
        'keypools': [
            {'type': 'small', 'settings': (), 'keys': 1,
             'chests': ('Castle Tower First Chest',)},
            {'type': 'small', 'settings': (), 'keys': 1,
             'chests': (
                 'Castle Tower First Chest', 'Castle Tower Seconds Chest')}],
        'link': {}
    },
    'Dark Palace': {
        'type': 'dungeon', 'map': 'dark', 'coord': (629, 241),
        'keypools': [
            {'type': 'small', 'settings': (), 'keys': 4,
             'chests': (
                 'Dark Palace First Key', 'Dark Palace Key With A View',
                 'Dark Palace Map', 'Dark Palace Big Key',
                 'Dark Palace Stalfos Key', 'Dark Palace Targetting Chest')},
            {'type': 'small', 'settings': (), 'keys': 2,
             'chests': (
                 'Dark Palace First Key', 'Dark Palace Key With A View',
                 'Dark Palace Map', 'Dark Palace Big Key',
                 'Dark Palace Stalfos Key', 'Dark Palace Targetting Chest',
                 'Dark Palace Treasury Left Chest',
                 'Dark Palace Treasury Right Chest','Dark Palace Compass')}],
        'link': {}
    },
    'Swamp Palace': {
        'type': 'dungeon', 'map': 'dark', 'coord': (304, 600),
        'keypools': [
            {'type': 'small', 'settings': (), 'keys': 1,
             'chests': ('Swamp Palace Key',)}],
        'link': {}
    },
    'Skull Woods': {
        'type': 'dungeon', 'map': 'dark', 'coord': (20, 22),
        'keypools': [],
        'link': {}
    },
    "Thieves' Town": {
        'type': 'dungeon', 'map': 'dark', 'coord': (77, 301),
        'keypools': [],
        'link': {}
    },
    'Ice Palace': {
        'type': 'dungeon', 'map': 'dark', 'coord': (523, 552),
        'keypools': [],
        'link': {}
    },
    'Misery Mire': {
        'type': 'dungeon', 'map': 'dark', 'coord': (43, 532),
        'keypools': [],
        'link': {}
    },
    'Turtle Rock': {
        'type': 'dungeon', 'map': 'dark', 'coord': (618, 33),
        'keypools': [
            {'type': 'small', 'settings': ('open',), 'keys': 1,
             'chests': (
                 'Turtle Rock Compass', 'Turtle Rock Map',
                 'Turtle Rock Map Key')},
            {'type': 'small', 'settings': ('open',), 'keys': 1,
             'chests': (
                 'Turtle Rock Compass', 'Turtle Rock Map',
                 'Turtle Rock Map Key', 'Turtle Rock Chain Chomp Key')},
            {'type': 'small', 'settings': ('open',), 'keys': 1,
             'chests': (
                 'Turtle Rock Compass', 'Turtle Rock Map',
                 'Turtle Rock Map Key', 'Turtle Rock Chain Chomp Key',
                 'Turtle Rock Big Key', 'Turtle Rock Treasure',
                 'Turtle Rock Switch Key')},
            {'type': 'small', 'settings': ('open',), 'keys': 1,
             'chests': (
                 'Turtle Rock Compass', 'Turtle Rock Map',
                 'Turtle Rock Map Key', 'Turtle Rock Chain Chomp Key',
                 'Turtle Rock Big Key', 'Turtle Rock Treasure',
                 'Turtle Rock Switch Key', 'Turtle Rock Treasury Chest 1',
                 'Turtle Rock Treasury Chest 2', 'Turtle Rock Treasury Chest 3',
                 'Turtle Rock Treasury Chest 4')},
            {'type': 'big', 'settings': ('open',),
             'chests': (
                 'Turtle Rock Compass', 'Turtle Rock Map',
                 'Turtle Rock Map Key', 'Turtle Rock Chain Chomp Key',
                 'Turtle Rock Big Key')}],
        'link': {}
    },
    "Ganon's Tower": {
        'type': 'dungeon', 'map': 'dark', 'coord': (367, 14),
        'keypools': [
            {'type': 'small', 'settings': (), 'keys': 2,
             'chests': (
                 "Ganon's Tower Torch Key",
                 "Ganon's Tower Stalfos Room Chest 1",
                 "Ganon's Tower Stalfos Room Chest 2",
                 "Ganon's Tower Stalfos Room Chest 3",
                 "Ganon's Tower Stalfos Room Chest 4",
                 "Ganon's Tower Trap Chest 1", "Ganon's Tower Trap Chest 2",
                 "Ganon's Tower Tile Room")},
            {'type': 'small', 'settings': (), 'keys': 1,
             'chests': (
                 "Ganon's Tower Torch Key",
                 "Ganon's Tower Stalfos Room Chest 1",
                 "Ganon's Tower Stalfos Room Chest 2",
                 "Ganon's Tower Stalfos Room Chest 3",
                 "Ganon's Tower Stalfos Room Chest 4",
                 "Ganon's Tower Map", "Ganon's Tower Trap Chest 1",
                 "Ganon's Tower Trap Chest 2", "Ganon's Tower Tile Room")},
            {'type': 'small', 'settings': (), 'keys': 1,
             'chests': (
                 "Ganon's Tower Torch Key",
                 "Ganon's Tower Stalfos Room Chest 1",
                 "Ganon's Tower Stalfos Room Chest 2",
                 "Ganon's Tower Stalfos Room Chest 3",
                 "Ganon's Tower Stalfos Room Chest 4",
                 "Ganon's Tower Map", "Ganon's Tower Winder Room Key",
                 "Ganon's Tower Trap Chest 1", "Ganon's Tower Trap Chest 2",
                 "Ganon's Tower Tile Room")}],
        'link': {}
    },
}



'''
Presets for configuration file
'''

import sys

from ..version import __version__ as version

__all__ = 'DEFAULT', 'OVERWRITE'


DEFAULT = {
    'gui': ('gui-tkinter', str),
    'font_size': ('14', int),
    'icon_size': ('1.0', float),
    'map_size': ('1.0', float),
    'autosave': ('autosave.json', str),
    'button_layout': ('buttons.json', str),
    'window_layout': ('windows.json', str),

    'entrance_randomiser': (True, bool),
    'world_state': ('Open', str),
    'glitch_mode': ('None', str),
    'item_placement': ('Advanced', str),
    'dungeon_items': ('Standard', str),
    'goal': ('Defeat Ganon', str),
    'swords': ('Randomised', str),
    'enemiser': (False, bool),

    'majoronly': (False, bool),
}


OVERWRITE = {
    '0.9.0': set(),
    '0.9.1': set(),
    '0.9.2': set(),
    '0.9.3': set(),
    '0.9.4': set(),
    '0.9.5': set(),
    '0.9.6': set(),
    '0.9.7': set(),
    '0.9.8': set(),
    '0.9.9': set(),
    '0.9.10': set()
}



'''
List of major item locations.
'''

__all__ = 'MAJORLOCATIONS',

MAJORLOCATIONS = {
    'East Palace Treasure', 'Pyramid Fairy', 'Castle Dungeons',
    'Waterfall Fairy', 'Swamp Palace Treasure', 'Mushroom', 'Witch Hut',
    'Skull Woods Treasure', 'Ice Cave', 'Bombos Stone', 'Ether Rock',
    'Catfish', "Link's House", 'Dark Palace Treasure', 'Ocarina Boy',
    'Haunted Grove', 'Bug Child', 'Library', 'Bottle Vendor',
    'Tavern Back Entrance', 'Under The Bridge', 'Locksmith',
    'Misery Mire Treasure', 'Cave of Byrna', "King's Tomb", 'Lost Man',
    "Sahasrahla's Reward", 'Desert Palace Treasure', "Thieves' Town Treasure",
    'King Zora', 'Mountain Tower Treasure', 'Secret Path',
    'Master Sword Pedestal', 'Blacksmiths', 'Turtle Rock Treasure',
    'Ice Palace Treasure', "Ganon's Tower Treasure", 'Mad Batter',
    'Sanctuary', 'East Palace Boss Item', 'Desert Palace Boss Item',
    'Mountain Tower Boss Item', 'Dark Palace Boss Item',
    'Swamp Palace Boss Item', 'Skull Woods Boss Item',
    "Thieves' Town Boss Item", 'Ice Palace Boss Item',
    'Misery Mire Boss Item', 'Turtle Rock Boss Item'}



import setuptools

setuptools.setup(
    name='z3-tracker',
    version='0.9.10',
    author='Feneg',
    description='Helper program for Link to the Past randomiser',
    url='https://www.github.com/feneg/z3-tracker',
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Topic :: Games/Entertainment',
        'Topic :: Utilities'],
    entry_points={
        'gui_scripts': (
            'z3-tracker = z3tracker.main:main',
            'z3tracker = z3tracker.main:main')}
    )



'''
Ocarina targets
'''

__all__ = 'LOCATIONS',


LOCATIONS = {
    'Ocarina': {
        'type': 'area',
        'link': {
            'West Lower Mount Hebra': [('nosettings', 'inverted')],
            'East Light World': [('nosettings', 'inverted')],
            'Kakariko': [('nosettings', 'inverted')],
            'South Light World': [('nosettings', 'inverted')],
            'East Light World': [('nosettings', 'inverted')],
            'Desert Portal (LW)': [('nosettings', 'inverted')],
            'Southern Shores': [('nosettings', 'inverted')],

            'West Lower Death Mountain': [('settings', 'inverted')],
            'River Shop Area': [('settings', 'inverted')],
            "Thieves' Town Surface": [('settings', 'inverted')],
            'South Dark World': [('settings', 'inverted')],
            'East Dark World': [('settings', 'inverted')],
            'Desert Portal (DW)': [('settings', 'inverted')],
            'Eastern Shores': [('settings', 'inverted')],
        }
    },

    'Light Bottle Ocarina': {
        'type': 'area',
        'link': {
            'West Lower Mount Hebra': [('and', [
                ('glitch', 'major'), ('item', 'bottle')])],
            'East Light World': [('and', [
                ('glitch', 'major'), ('item', 'bottle')])],
            'Kakariko': [('and', [
                ('glitch', 'major'), ('item', 'bottle')])],
            'South Light World': [('and', [
                ('glitch', 'major'), ('item', 'bottle')])],
            'East Light World': [('and', [
                ('glitch', 'major'), ('item', 'bottle')])],
            'Desert Portal (LW)': [('and', [
                ('glitch', 'major'), ('item', 'bottle')])],
            'Southern Shores': [('and', [
                ('glitch', 'major'), ('item', 'bottle')])]}
    },

    'Dark Bottle Ocarina': {
        'type': 'area',
        'link': {
            'West Lower Death Mountain': [('and', [
                ('glitch', 'major'), ('item', 'bottle')])],
            'River Shop Area': [('and', [
                ('glitch', 'major'), ('item', 'bottle')])],
            "Thieves' Town Surface": [('and', [
                ('glitch', 'major'), ('item', 'bottle')])],
            'South Dark World': [('and', [
                ('glitch', 'major'), ('item', 'bottle')])],
            'East Dark World': [('and', [
                ('glitch', 'major'), ('item', 'bottle')])],
            'Desert Portal (DW)': [('and', [
                ('glitch', 'major'), ('item', 'bottle')])],
            'Eastern Shores': [('and', [
                ('glitch', 'major'), ('item', 'bottle')])]}
    },
}



'''
Interface to configuration system
'''

import json
import os
import sys
import typing

from ..version import __version__ as version

from .default import DEFAULT, OVERWRITE


__all__ = 'Config', 'config_directory'


class ConfigOption(dict):
    '''
    Single entry info for configuration file.

    Instance variables:
        name: name of the config option
        default: default value
    '''

    def __init__(
            self, name: str,
            valtype: typing.Callable[[typing.Any], typing.Any],
            default: typing.Any):
        '''
        Args:
            name: name of the config option
            valtype: type of the config option
            default: default value
        '''

        super().__init__()
        self.name = name
        self.valtype = valtype
        self.default = default


class Config(dict):
    '''
    Program configuration.

    Instance variables:
        self: program configuration {'option': value}
        filename: full path to config filename
    '''

    def __init__(self):
        super().__init__()

        self.filename = _config_filename()

        try:
            fid = open(self.filename, 'r')
        except FileNotFoundError:
            imported = None
        else:
            imported = fid.read()
            fid.close()

        try:
            imp_dict = {} if imported is None else json.loads(imported)
        except json.decoder.JSONDecodeError:
            imp_dict = {}

        for entry in imp_dict:
            self._insert_entry(entry, imp_dict[entry])
        self._check_missing()

        self['version'] = version

    def _write_config(self) -> None:
        '''
        Write current configuration to file.
        '''

        self['version'] = version
        with open(self.filename, 'w') as fid:
            json.dump(self, fid)

    def _insert_entry(self, entry_name: str, entry_value: typing.Any) -> None:
        '''
        Insert imported entry.

        Args:
            entry: config file entry read from file
        '''

        if entry_name in DEFAULT:
            self[entry_name] = DEFAULT[entry_name][1](entry_value)

    def _check_missing(self) -> None:
        '''
        Insert missing entries into configuration.
        '''

        changed = False
        if 'version' not in self or self['version'] != version:
            for entry in OVERWRITE[version]:
                self[entry] = DEFAULT[entry][1](DEFAULT[entry][0])
                changed = True
        for entry in DEFAULT:
            if entry not in self:
                self[entry] = DEFAULT[entry][1](DEFAULT[entry][0])
                changed = True
        if changed:
            self._write_config()

    def update(self, entry: str, data: object) -> None:
        '''
        Update config entry with new data.

        Args:
            entry: config entry
            data: new data
        '''

        self[entry] = data
        self._write_config()


def config_directory() -> str:
    '''
    Return configuration directory.

    This will create said directory if it doesn't exist.

    Returns:
        str: full path to configuration directory
    '''

    if sys.platform.startswith('win32'):
        configdir = os.path.join(os.getenv('LOCALAPPDATA'), 'z3-tracker')
    else:
        configdir = os.path.expanduser('~/.z3-tracker')
    if not os.path.isdir(configdir):
        os.mkdir(configdir)
    return configdir


def _config_filename() -> str:
    '''
    Return config file name.

    This will create the config directory if it doesn't exist. The config file
    itself, however, won't.

    Returns:
        str: fill path to config file
    '''

    return os.path.join(config_directory(), 'config.json')
